"""
Pipeline modules for estimating detection limits.
"""

import os
import math
import time
import warnings
import multiprocessing as mp

from typing import List, Optional, Tuple

import numpy as np

from scipy.interpolate import griddata
from typeguard import typechecked

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.image import create_mask
from pynpoint.util.limits import contrast_limit
from pynpoint.util.module import progress
from pynpoint.util.psf import pca_psf_subtraction
from pynpoint.util.residuals import combine_residuals


class ContrastCurveModule(ProcessingModule):
    """
    Pipeline module to calculate contrast limits for a given sigma level or false positive
    fraction, with a correction for small sample statistics. Positions are processed in
    parallel if ``CPU`` is set to a value larger than 1 in the configuration file.
    """

    __author__ = 'Tomas Stolker, Jasper Jonker, Benedikt Schmidhuber'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 psf_in_tag: str,
                 contrast_out_tag: str,
                 separation: Tuple[float, float, float] = (0.1, 1., 0.01),
                 angle: Tuple[float, float, float] = (0., 360., 60.),
                 threshold: Tuple[str, float] = ('sigma', 5.),
                 psf_scaling: float = 1.,
                 aperture: float = 0.05,
                 pca_number: int = 20,
                 cent_size: Optional[float] = None,
                 edge_size: Optional[float] = None,
                 extra_rot: float = 0.,
                 residuals: str = 'mean',
                 snr_inject: float = 100.,
                 **kwargs: float) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry that contains the stack with images.
        psf_in_tag : str
            Tag of the database entry that contains the reference PSF that is used as fake planet.
            Can be either a single image or a stack of images equal in size to *image_in_tag*.
        contrast_out_tag : str
            Tag of the database entry that contains the separation, azimuthally averaged contrast
            limits, the azimuthal variance of the contrast limits, and the threshold of the false
            positive fraction associated with sigma.
        separation : tuple(float, float, float)
            Range of separations (arcsec) where the contrast is calculated. Should be specified as
            (lower limit, upper limit, step size). Apertures that fall within the mask radius or
            beyond the image size are removed.
        angle : tuple(float, float, float)
            Range of position angles (deg) where the contrast is calculated. Should be specified as
            (lower limit, upper limit, step size), measured counterclockwise with respect to the
            vertical image axis, i.e. East of North.
        threshold : tuple(str, float)
            Detection threshold for the contrast curve, either in terms of 'sigma' or the false
            positive fraction (FPF). The value is a tuple, for example provided as ('sigma', 5.)
            or ('fpf', 1e-6). Note that when sigma is fixed, the false positive fraction will
            change with separation. Also, sigma only corresponds to the standard deviation of a
            normal distribution at large separations (i.e., large number of samples).
        psf_scaling : float
            Additional scaling factor of the planet flux (e.g., to correct for a neutral density
            filter). Should have a positive value.
        aperture : float
            Aperture radius (arcsec).
        pca_number : int
            Number of principal components used for the PSF subtraction.
        cent_size : float, None
            Central mask radius (arcsec). No mask is used when set to None.
        edge_size : float, None
            Outer edge radius (arcsec) beyond which pixels are masked. No outer mask is used when
            set to None. If the value is larger than half the image size then it will be set to
            half the image size.
        extra_rot : float
            Additional rotation angle of the images in clockwise direction (deg).
        residuals : str
            Method used for combining the residuals ('mean', 'median', 'weighted', or 'clipped').
        snr_inject : float
            Signal-to-noise ratio of the injected planet signal that is used to measure the amount
            of self-subtraction.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        if 'sigma' in kwargs:
            warnings.warn('The \'sigma\' parameter has been deprecated. Please use the '
                          '\'threshold\' parameter instead.', DeprecationWarning)

        if 'norm' in kwargs:
            warnings.warn('The \'norm\' parameter has been deprecated. It is not recommended to '
                          'normalize the images before PSF subtraction.', DeprecationWarning)

        if 'accuracy' in kwargs:
            warnings.warn('The \'accuracy\' parameter has been deprecated. The parameter is no '
                          'longer required.', DeprecationWarning)

        if 'magnitude' in kwargs:
            warnings.warn('The \'magnitude\' parameter has been deprecated. The parameter is no '
                          'longer required.', DeprecationWarning)

        if 'ignore' in kwargs:
            warnings.warn('The \'ignore\' parameter has been deprecated. The parameter is no '
                          'longer required.', DeprecationWarning)

        self.m_image_in_port = self.add_input_port(image_in_tag)

        if psf_in_tag == image_in_tag:
            self.m_psf_in_port = self.m_image_in_port
        else:
            self.m_psf_in_port = self.add_input_port(psf_in_tag)

        self.m_contrast_out_port = self.add_output_port(contrast_out_tag)

        self.m_separation = separation
        self.m_angle = angle
        self.m_psf_scaling = psf_scaling
        self.m_threshold = threshold
        self.m_aperture = aperture
        self.m_pca_number = pca_number
        self.m_cent_size = cent_size
        self.m_edge_size = edge_size
        self.m_extra_rot = extra_rot
        self.m_residuals = residuals
        self.m_snr_inject = snr_inject

        if self.m_angle[0] < 0. or self.m_angle[0] > 360. or self.m_angle[1] < 0. or \
           self.m_angle[1] > 360. or self.m_angle[2] < 0. or self.m_angle[2] > 360.:

            raise ValueError('The angular positions of the fake planets should lie between '
                             '0 deg and 360 deg.')

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. An artificial planet is injected (based on the noise level) at a
        given separation and position angle. The amount of self-subtraction is then determined and
        the contrast limit is calculated for a given sigma level or false positive fraction. A
        correction for small sample statistics is applied for both cases. Note that if the sigma
        level is fixed, the false positive fraction changes with separation, following the
        Student's t-distribution (see Mawet et al. 2014 for details).

        Returns
        -------
        NoneType
            None
        """

        images = self.m_image_in_port.get_all()
        psf = self.m_psf_in_port.get_all()

        if psf.shape[0] != 1 and psf.shape[0] != images.shape[0]:
            raise ValueError(f'The number of frames in psf_in_tag {psf.shape} does not match with '
                             f'the number of frames in image_in_tag {images.shape}. The '
                             f'DerotateAndStackModule can be used to average the PSF frames '
                             f'(without derotating) before applying the ContrastCurveModule.')

        cpu = self._m_config_port.get_attribute('CPU')
        working_place = self._m_config_port.get_attribute('WORKING_PLACE')

        parang = self.m_image_in_port.get_attribute('PARANG')
        pixscale = self.m_image_in_port.get_attribute('PIXSCALE')

        self.m_image_in_port.close_port()
        self.m_psf_in_port.close_port()

        if self.m_cent_size is not None:
            self.m_cent_size /= pixscale

        if self.m_edge_size is not None:
            self.m_edge_size /= pixscale

        self.m_aperture /= pixscale

        pos_r = np.arange(self.m_separation[0]/pixscale,
                          self.m_separation[1]/pixscale,
                          self.m_separation[2]/pixscale)

        pos_t = np.arange(self.m_angle[0]+self.m_extra_rot,
                          self.m_angle[1]+self.m_extra_rot,
                          self.m_angle[2])

        if self.m_cent_size is None:
            index_del = np.argwhere(pos_r-self.m_aperture <= 0.)
        else:
            index_del = np.argwhere(pos_r-self.m_aperture <= self.m_cent_size)

        pos_r = np.delete(pos_r, index_del)

        if self.m_edge_size is None or self.m_edge_size > images.shape[1]/2.:
            index_del = np.argwhere(pos_r+self.m_aperture >= images.shape[1]/2.)
        else:
            index_del = np.argwhere(pos_r+self.m_aperture >= self.m_edge_size)

        pos_r = np.delete(pos_r, index_del)

        positions = []
        for sep in pos_r:
            for ang in pos_t:
                positions.append((sep, ang))

        result = []
        async_results = []

        # Create temporary files
        tmp_im_str = os.path.join(working_place, 'tmp_images.npy')
        tmp_psf_str = os.path.join(working_place, 'tmp_psf.npy')

        np.save(tmp_im_str, images)
        np.save(tmp_psf_str, psf)

        mask = create_mask(images.shape[-2:], (self.m_cent_size, self.m_edge_size))

        _, im_res = pca_psf_subtraction(images=images*mask,
                                        angles=-1.*parang+self.m_extra_rot,
                                        pca_number=self.m_pca_number)

        noise = combine_residuals(method=self.m_residuals, res_rot=im_res)

        pool = mp.Pool(cpu)

        for pos in positions:
            async_results.append(pool.apply_async(contrast_limit,
                                                  args=(tmp_im_str,
                                                        tmp_psf_str,
                                                        noise,
                                                        mask,
                                                        parang,
                                                        self.m_psf_scaling,
                                                        self.m_extra_rot,
                                                        self.m_pca_number,
                                                        self.m_threshold,
                                                        self.m_aperture,
                                                        self.m_residuals,
                                                        self.m_snr_inject,
                                                        pos)))

        pool.close()

        start_time = time.time()

        # wait for all processes to finish
        while mp.active_children():
            # number of finished processes
            nfinished = sum([i.ready() for i in async_results])

            progress(nfinished, len(positions), 'Calculating detection limits...', start_time)

            # check if new processes have finished every 5 seconds
            time.sleep(5)

        if nfinished != len(positions):
            print('\r                                                      ')
            print('\rCalculating detection limits... [DONE]')

        # get the results for every async_result object
        for item in async_results:
            result.append(item.get())

        pool.terminate()

        os.remove(tmp_im_str)
        os.remove(tmp_psf_str)

        result = np.asarray(result)

        # Sort the results first by separation and then by angle
        indices = np.lexsort((result[:, 1], result[:, 0]))
        result = result[indices]

        result = result.reshape((pos_r.size, pos_t.size, 4))
        result[np.isinf(result)] = np.nan

        mag_mean = np.nanmean(result, axis=1)[:, 2]
        mag_var = np.nanvar(result, axis=1)[:, 2]
        res_fpf = result[:, 0, 3]

        limits = np.column_stack((pos_r*pixscale, mag_mean, mag_var, res_fpf))

        self.m_image_in_port._check_status_and_activate()
        self.m_contrast_out_port._check_status_and_activate()

        self.m_contrast_out_port.set_all(limits, data_dim=2)

        history = f'{self.m_threshold[0]} = {self.m_threshold[1]}'
        self.m_contrast_out_port.add_history('ContrastCurveModule', history)
        self.m_contrast_out_port.copy_attributes(self.m_image_in_port)
        self.m_contrast_out_port.close_port()


class MassLimitsModule(ProcessingModule):
    """
    Pipeline module to calculate mass limits from the contrast limits and any isochrones model grid
    downloaded from https://phoenix.ens-lyon.fr/Grids/.
    """

    __author__ = 'Benedikt Schmidhuber, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 contrast_in_tag: str,
                 mass_out_tag: str,
                 model_file: str,
                 star_prop: dict,
                 instr_filter: str = 'L\'') -> None:

        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        contrast_in_tag : str
            Tag of the database entry that contains the contrast curve data, as computed with the
            :class:`~pynpoint.processing.limits.ContrastCurveModule`.
        mass_out_tag : str
            Tag of the database entry with the output data containing the separation, the mass
            limits, and the upper and lower one sigma deviation as calculated for the azimuthal
            variance on the contrast limits.
        model_file: str
            Path to the file containing the model data. Must be in the same format as the
            grids found on https://phoenix.ens-lyon.fr/Grids/. Any of the isochrones files from
            this website can be used.
        star_prop : dict
            Dictionary containing host star properties. Must have the following keys:

                - ``magnitude`` - Apparent magnitude, in the same band as the `instr_filter`.
                - ``distance`` - Distance in parsec.
                - ``age`` - Age of the system in the Myr.

        instr_filter: str
            Instrument filter in the same format as listed in the `model_file`.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_star_age = star_prop['age']/1000.  # [Myr]
        self.m_star_abs = star_prop['magnitude'] - 5.*math.log10(star_prop['distance']/10.)

        self.m_instr_filter = instr_filter
        self.m_model_file = model_file

        if not os.path.exists(self.m_model_file):
            raise ValueError('The path does not appear to be an existing file. Please check the'
                             'path. If you are unsure about the path, pass the absolute path to the'
                             'model file.')

        self.m_contrast_in_port = self.add_input_port(contrast_in_tag)
        self.m_mass_out_port = self.add_output_port(mass_out_tag)

    @staticmethod
    @typechecked
    def read_model(model_file_path: str) -> Tuple[List[float], List[np.ndarray], List[str]]:
        """
        Reads the data from the model file and structures it. Returns an array of available model
        ages and a list of model data for each age.

        Parameters
        -------
        model_file: str
            Path to the file containing the model data.

        Returns
        -------
        list(float)
            List with all the ages from the model grid.
        list(np.ndarray)
            List with all the isochrone data, so the length is the same as the number of ages.
        list(str)
            List with all the column names from the model grid.
        """

        # read in all the data, selecting out empty lines or '---' lines
        data = []
        with open(model_file_path) as file:
            for line in file:
                if ('---' in line) or line == '\n':
                    continue
                else:
                    data += [list(filter(None, line.rstrip().split(' ')))]

        # initialize list of ages
        ages = []
        # initialize the header
        header = []
        # initialize a new data list, where the data is separated by age
        isochrones = []

        k = -1
        for _line in data:
            if '(Gyr)' in _line:
                # get time line
                ages += [float(_line[-1])]
                isochrones += [[]]
                k += 1

            elif 'lg(g)' in _line:
                # get header line
                header = ['M/Ms', 'Teff(K)'] + _line[1:]

            else:
                # save the data
                isochrones[k] += [_line]

        for index, _ in enumerate(isochrones):
            isochrones[index] = np.array(isochrones[index], dtype=float)

        return ages, isochrones, header

    @staticmethod
    @typechecked
    def interpolate_model(age_eval: np.ndarray,
                          mag_eval: np.ndarray,
                          filter_index: int,
                          model_age: List[float],
                          model_data: List[np.ndarray]) -> np.ndarray:
        """
        Interpolates the grid based model data.

        Parameters
        ----------
        age_eval : np.ndarray
            Age at which the system is evaluated. Must be of the same shape as `mag_eval`.
        mag_eval : np.ndarray
            Absolute magnitude for which the system is evaluated. Must be of the same shape as
            `age_eval`.
        filter_index: int
            Column index where the filter is located.
        model_age: list(float)
            List of ages which are given by the model.
        model_data: list(np.ndarray)
            List of arrays containing the model data.

        Returns
        -------
        griddata : np.ndarray
            Interpolated values for the given evaluation points (age_eval, mag_eval). Has the
            same shape as age_eval and mag_eval.
        """

        grid_points = np.array([])
        grid_values = np.array([])

        # create array of available points
        for age_index, age_item in enumerate(model_age):
            iso_mag = model_data[age_index][:, filter_index]
            iso_age = np.ones_like(iso_mag) * age_item
            iso_mass = model_data[age_index][:, 0]

            grid_points = np.append(grid_points, np.column_stack((iso_age, iso_mag)))
            grid_values = np.append(grid_values, iso_mass)

        grid_points = grid_points.reshape(-1, 2)
        interp = np.column_stack((age_eval, mag_eval))

        return griddata(grid_points, grid_values, interp, method='cubic', rescale=True)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Calculates the mass limits from the contrast limits (as
        calculated with the :class:`~pynpoint.processing.limits.ContrastCurveModule`) and the
        isochrones of an evolutionary model. The age and the absolute magnitude of the isochrones
        are linearly interpolated such that the mass limits can be calculated for a given contrast
        limits  (which is converted in an absolute magnitude with the apparent magnitude and
        distance of the central star).

        Returns
        -------
        NoneType
            None
        """

        model_age, model_data, model_header = self.read_model(self.m_model_file)

        assert self.m_instr_filter in model_header, 'The selected filter was not found in the ' \
                                                    'list of available filters from the model.'

        # find the column index of the filter
        # simple argwhere gives empty list?!
        filter_index = np.argwhere([self.m_instr_filter == j for j in model_header])[0]
        filter_index = int(filter_index)

        contrast_data = self.m_contrast_in_port.get_all()

        separation = contrast_data[:, 0]
        contrast = contrast_data[:, 1]
        contrast_std = np.sqrt(contrast_data[:, 2])

        age_eval = self.m_star_age*np.ones_like(contrast)
        mag_eval = self.m_star_abs+contrast

        print('Interpolating isochrones...', end='')

        mass = self.interpolate_model(age_eval=age_eval,
                                      mag_eval=mag_eval,
                                      filter_index=filter_index,
                                      model_age=model_age,
                                      model_data=model_data)

        mass_upper = self.interpolate_model(age_eval=age_eval,
                                            mag_eval=mag_eval-contrast_std,
                                            filter_index=filter_index,
                                            model_age=model_age,
                                            model_data=model_data) - mass

        mass_lower = self.interpolate_model(age_eval=age_eval,
                                            mag_eval=mag_eval+contrast_std,
                                            filter_index=filter_index,
                                            model_age=model_age,
                                            model_data=model_data) - mass

        mass_limits = np.column_stack((separation, mass, mass_upper, mass_lower))
        self.m_mass_out_port.set_all(mass_limits, data_dim=2)

        print(' [DONE]')

        history = f'filter = {self.m_instr_filter}'
        self.m_mass_out_port.add_history('MassLimitsModule', history)
        self.m_mass_out_port.copy_attributes(self.m_contrast_in_port)
        self.m_mass_out_port.close_port()



"""
Functions to write selected data and attributes to the database.
"""

import time

from typing import Optional, Union

import numpy as np

from typeguard import typechecked

from pynpoint.core.dataio import InputPort, OutputPort
from pynpoint.util.module import progress, memory_frames


@typechecked
def write_selected_data(memory: Union[int, np.int64],
                        indices: np.ndarray,
                        image_in_port: InputPort,
                        selected_out_port: Optional[OutputPort],
                        removed_out_port: Optional[OutputPort]) -> None:
    """
    Function to write the selected and removed data.

    Parameters
    ----------
    memory : int
        Number of images that is simultaneously loaded into the memory.
    indices : numpy.ndarray
        Image indices that will be removed.
    image_in_port : pynpoint.core.dataio.InputPort
        Port to the input images.
    selected_out_port : pynpoint.core.dataio.OutputPort, None
        Port to store the selected images. No data is written if set to None.
    removed_out_port : pynpoint.core.dataio.OutputPort, None
        Port to store the removed images. No data is written if set to None.

    Returns
    -------
    NoneType
        None
    """

    nimages = image_in_port.get_shape()[0]
    frames = memory_frames(memory, nimages)

    if memory == 0 or memory >= nimages:
        memory = nimages

    start_time = time.time()

    for i, _ in enumerate(frames[:-1]):
        progress(i, len(frames[:-1]), 'Writing selected data...', start_time)

        images = image_in_port[frames[i]:frames[i+1], ]

        subset_del = np.where(np.logical_and(indices >= frames[i], indices < frames[i+1]))[0]
        index_del = indices[subset_del] % memory

        index_sel = np.ones(images.shape[0], bool)
        index_sel[index_del] = False

        if selected_out_port is not None and index_sel.size > 0:
            selected_out_port.append(images[index_sel])

        if removed_out_port is not None and index_del.size > 0:
            removed_out_port.append(images[index_del])


@typechecked
def write_selected_attributes(indices: np.ndarray,
                              image_in_port: InputPort,
                              selected_out_port: Optional[OutputPort],
                              removed_out_port: Optional[OutputPort],
                              module_type: str,
                              history: str) -> None:
    """
    Function to write the attributes of the selected and removed data.

    Parameters
    ----------
    indices : numpy.ndarray
        Image indices that will be removed.
    image_in_port : pynpoint.core.dataio.InputPort
        Port to the input data.
    selected_out_port : pynpoint.core.dataio.OutputPort, None
        Port to store the attributes of the selected images. Not written if set to None.
    removed_out_port : pynpoint.core.dataio.OutputPort, None
        Port to store the attributes of the removed images. Not written if set to None.
    module_type : str
    history : str

    Returns
    -------
    NoneType
        None
    """

    if selected_out_port is not None:
        # First copy the existing attributes to the selected_out_port
        selected_out_port.copy_attributes(image_in_port)
        selected_out_port.add_history(module_type, history)

    if removed_out_port is not None:
        # First copy the existing attributes to the removed_out_port
        removed_out_port.copy_attributes(image_in_port)
        removed_out_port.add_history(module_type, history)

    non_static = image_in_port.get_all_non_static_attributes()

    index_sel = np.ones(image_in_port.get_shape()[0], bool)
    index_sel[indices] = False

    for i, attr_item in enumerate(non_static):
        values = image_in_port.get_attribute(attr_item)

        if values.shape[0] == image_in_port.get_shape()[0]:

            if selected_out_port is not None and index_sel.size > 0:
                selected_out_port.add_attribute(attr_item, values[index_sel], static=False)

            if removed_out_port is not None and indices.size > 0:
                removed_out_port.add_attribute(attr_item, values[indices], static=False)

    if 'NFRAMES' in non_static:
        nframes = image_in_port.get_attribute('NFRAMES')

        nframes_sel = np.zeros(nframes.shape, dtype=int)
        nframes_del = np.zeros(nframes.shape, dtype=int)

        for i, frames in enumerate(nframes):
            if indices.size == 0:
                nframes_sel[i] = frames
                nframes_del[i] = 0

            else:
                sum_n = np.sum(nframes[:i])
                index_del = np.where(np.logical_and(indices >= sum_n, indices < sum_n+frames))[0]

                nframes_sel[i] = frames - index_del.size
                nframes_del[i] = index_del.size

        if selected_out_port is not None:
            selected_out_port.add_attribute('NFRAMES', nframes_sel, static=False)

        if removed_out_port is not None:
            removed_out_port.add_attribute('NFRAMES', nframes_del, static=False)



"""
Module to obtain information about the implemented attributes.
"""

from typing import Dict, Union

from typeguard import typechecked


@typechecked
def get_attributes() -> Dict[str, Dict[str, Union[str, float, int, None]]]:
    """
    Function to get a dictionary with all attributes.

    Returns
    -------
    dict
        Attribute information.
    """

    attr = {'PIXSCALE': {'attribute': 'static',
                         'config': 'settings',
                         'value': 0.027,
                         'type': 'float'},
            'MEMORY': {'attribute': 'static',
                       'config': 'settings',
                       'value': 1000,
                       'type': 'int'},
            'CPU': {'attribute': 'static',
                    'config': 'settings',
                    'value': 1,
                    'type': 'int'},
            'INSTRUMENT': {'attribute': 'static',
                           'config': 'header',
                           'value': 'INSTRUME',
                           'type': 'str'},
            'NFRAMES': {'attribute': 'non-static',
                        'config': 'header',
                        'value': 'NAXIS3',
                        'type': 'int'},
            'EXP_NO': {'attribute': 'non-static',
                       'config': 'header',
                       'value': 'ESO DET EXP NO',
                       'type': 'int'},
            'DIT': {'attribute': 'static',
                    'config': 'header',
                    'value': 'ESO DET DIT',
                    'type': 'int'},
            'NDIT': {'attribute': 'non-static',
                     'config': 'header',
                     'value': 'ESO DET NDIT',
                     'type': 'int'},
            'PARANG_START': {'attribute': 'non-static',
                             'config': 'header',
                             'value': 'ESO ADA POSANG',
                             'type': 'float'},
            'PARANG_END': {'attribute': 'non-static',
                           'config': 'header',
                           'value': 'ESO ADA POSANG END',
                           'type': 'float'},
            'DITHER_X': {'attribute': 'non-static',
                         'config': 'header',
                         'value': 'ESO SEQ CUMOFFSETX',
                         'type': 'float'},
            'DITHER_Y': {'attribute': 'non-static',
                         'config': 'header',
                         'value': 'ESO SEQ CUMOFFSETY',
                         'type': 'float'},
            'PUPIL': {'attribute': 'non-static',
                      'config': 'header',
                      'value': 'ESO ADA PUPILPOS',
                      'type': 'float'},
            'DATE': {'attribute': 'non-static',
                     'config': 'header',
                     'value': 'DATE-OBS',
                     'type': 'str'},
            'LATITUDE': {'attribute': 'static',
                         'config': 'header',
                         'value': 'ESO TEL GEOLAT',
                         'type': 'float'},
            'LONGITUDE': {'attribute': 'static',
                          'config': 'header',
                          'value': 'ESO TEL GEOLON',
                          'type': 'float'},
            'RA': {'attribute': 'non-static',
                   'config': 'header',
                   'value': 'RA',
                   'type': 'float'},
            'DEC': {'attribute': 'non-static',
                    'config': 'header',
                    'value': 'DEC',
                    'type': 'float'},
            'PARANG': {'attribute': 'non-static',
                       'config': 'header',
                       'value': 'None',
                       'type': 'float'},
            'WAVELENGTH': {'attribute': 'non-static',
                           'config': 'header',
                           'value': 'None',
                           'type': 'float'},
            'STAR_POSITION': {'attribute': 'non-static',
                              'config': None,
                              'value': None,
                              'type': 'float'},
            'INDEX': {'attribute': 'non-static',
                      'config': None,
                      'value': None,
                      'type': 'int'},
            'FILES': {'attribute': 'non-static',
                      'config': None,
                      'value': None,
                      'type': 'str'}}

    return attr



import os

import pytest
import numpy as np

from astropy.io import fits

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.nearreading import NearReadingModule
from pynpoint.util.tests import create_config, create_near_data, remove_test_data


class TestNearReading:

    def setup_class(self) -> None:

        self.limit = 1e-8
        self.test_dir = os.path.dirname(__file__) + '/'
        self.fitsfile = self.test_dir + 'near/images_0.fits'

        create_near_data(path=self.test_dir + 'near')
        create_config(self.test_dir + 'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

        self.pipeline.set_attribute('config', 'NFRAMES', 'ESO DET CHOP NCYCLES', static=True)
        self.pipeline.set_attribute('config', 'EXP_NO', 'ESO TPL EXPNO', static=True)
        self.pipeline.set_attribute('config', 'NDIT', 'None', static=True)
        self.pipeline.set_attribute('config', 'PARANG_START', 'None', static=True)
        self.pipeline.set_attribute('config', 'PARANG_END', 'None', static=True)
        self.pipeline.set_attribute('config', 'DITHER_X', 'None', static=True)
        self.pipeline.set_attribute('config', 'DITHER_Y', 'None', static=True)
        self.pipeline.set_attribute('config', 'PIXSCALE', 0.045, static=True)
        self.pipeline.set_attribute('config', 'MEMORY', 100, static=True)

        self.positions = ('chopa', 'chopb')

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['near'])

    def test_near_read(self) -> None:

        module = NearReadingModule(name_in='read1a',
                                   input_dir=self.test_dir+'near',
                                   chopa_out_tag=self.positions[0],
                                   chopb_out_tag=self.positions[1])

        self.pipeline.add_module(module)
        self.pipeline.run_module('read1a')

        for item in self.positions:

            data = self.pipeline.get_data(item)
            assert np.mean(data) == pytest.approx(0.060582854, rel=self.limit, abs=0.)
            assert data.shape == (10, 10, 10)

    def test_near_subtract_crop_mean(self) -> None:

        module = NearReadingModule(name_in='read1b',
                                   input_dir=self.test_dir+'near',
                                   chopa_out_tag=self.positions[0],
                                   chopb_out_tag=self.positions[1],
                                   subtract=True,
                                   crop=(None, None, 0.3),
                                   combine='mean')

        self.pipeline.add_module(module)
        self.pipeline.run_module('read1b')

        data = self.pipeline.get_data(self.positions[0])
        assert np.mean(data) == pytest.approx(0., rel=self.limit, abs=0.)
        assert data.shape == (2, 7, 7)

        data = self.pipeline.get_data(self.positions[1])
        assert np.mean(data) == pytest.approx(0., rel=self.limit, abs=0.)
        assert data.shape == (2, 7, 7)

    def test_near_median(self) -> None:

        module = NearReadingModule(name_in='read1c',
                                   input_dir=self.test_dir+'near',
                                   chopa_out_tag=self.positions[0],
                                   chopb_out_tag=self.positions[1],
                                   combine='median')

        self.pipeline.add_module(module)
        self.pipeline.run_module('read1c')

        data = self.pipeline.get_data(self.positions[0])
        assert np.mean(data) == pytest.approx(0.060582854, rel=self.limit, abs=0.)
        assert data.shape == (2, 10, 10)

        data = self.pipeline.get_data(self.positions[1])
        assert np.mean(data) == pytest.approx(0.060582854, rel=self.limit, abs=0.)
        assert data.shape == (2, 10, 10)

    def test_static_not_found(self) -> None:

        self.pipeline.set_attribute('config', 'DIT', 'Test', static=True)

        module = NearReadingModule(name_in='read2',
                                   input_dir=self.test_dir+'near',
                                   chopa_out_tag=self.positions[0],
                                   chopb_out_tag=self.positions[1])

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('read2')

        assert len(warning) == 4

        for item in warning:
            assert item.message.args[0] == 'Static attribute DIT (=Test) not found in the FITS ' \
                                           'header.'

        self.pipeline.set_attribute('config', 'DIT', 'ESO DET SEQ1 DIT', static=True)

    def test_nonstatic_not_found(self) -> None:

        self.pipeline.set_attribute('config', 'NDIT', 'Test', static=True)

        module = NearReadingModule(name_in='read3',
                                   input_dir=self.test_dir+'near',
                                   chopa_out_tag=self.positions[0],
                                   chopb_out_tag=self.positions[1])

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('read3')

        assert len(warning) == 4

        for item in warning:
            assert item.message.args[0] == 'Non-static attribute NDIT (=Test) not found in the ' \
                                           'FITS header.'

        self.pipeline.set_attribute('config', 'NDIT', 'None', static=True)

    def test_check_header(self) -> None:

        with fits.open(self.fitsfile) as hdulist:
            hdulist[0].header['ESO DET CHOP ST'] = 'F'
            hdulist[0].header['ESO DET CHOP CYCSKIP'] = 1
            hdulist[0].header['ESO DET CHOP CYCSUM'] = 'T'
            hdulist.writeto(self.fitsfile, overwrite=True)

        module = NearReadingModule(name_in='read4',
                                   input_dir=self.test_dir+'near',
                                   chopa_out_tag=self.positions[0],
                                   chopb_out_tag=self.positions[1])

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('read4')

        assert len(warning) == 3

        assert warning[0].message.args[0] == 'Dataset was obtained without chopping.'
        assert warning[1].message.args[0] == 'Chop cycles (1) have been skipped.'
        assert warning[2].message.args[0] == 'FITS file contains averaged images.'

        with fits.open(self.fitsfile) as hdulist:
            hdulist[0].header['ESO DET CHOP ST'] = 'T'
            hdulist[0].header['ESO DET CHOP CYCSKIP'] = 0
            hdulist[0].header['ESO DET CHOP CYCSUM'] = 'F'
            hdulist.writeto(self.fitsfile, overwrite=True)

    def test_frame_type_invalid(self) -> None:

        with fits.open(self.fitsfile) as hdulist:
            hdulist[10].header['ESO DET FRAM TYPE'] = 'Test'
            hdulist.writeto(self.fitsfile, overwrite=True)

        module = NearReadingModule(name_in='read5',
                                   input_dir=self.test_dir+'near',
                                   chopa_out_tag=self.positions[0],
                                   chopb_out_tag=self.positions[1])

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('read5')

        assert str(error.value) == 'Frame type (Test) not a valid value. Expecting HCYCLE1 or ' \
                                   'HCYCLE2 as value for ESO DET FRAM TYPE.'

    def test_frame_type_missing(self) -> None:

        with fits.open(self.fitsfile) as hdulist:
            hdulist[10].header.remove('ESO DET FRAM TYPE')
            hdulist.writeto(self.fitsfile, overwrite=True)

        module = NearReadingModule(name_in='read6',
                                   input_dir=self.test_dir+'near',
                                   chopa_out_tag=self.positions[0],
                                   chopb_out_tag=self.positions[1])

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('read6')

        assert str(error.value) == 'Frame type not found in the FITS header. Image number: 9.'

    def test_same_cycle(self) -> None:

        with fits.open(self.fitsfile) as hdulist:

            with pytest.warns(UserWarning) as warning:
                hdulist[10].header['ESO DET FRAM TYPE'] = 'HCYCLE1'

            assert len(warning) == 1

            assert warning[0].message.args[0] == 'Keyword name \'ESO DET FRAM TYPE\' is greater ' \
                                                 'than 8 characters or contains characters not ' \
                                                 'allowed by the FITS standard; a HIERARCH card ' \
                                                 'will be created.'

            hdulist.writeto(self.fitsfile, overwrite=True)

        module = NearReadingModule(name_in='read7',
                                   input_dir=self.test_dir+'near',
                                   chopa_out_tag=self.positions[0],
                                   chopb_out_tag=self.positions[1])

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('read7')

        assert len(warning) == 2

        assert warning[0].message.args[0] == 'Previous and current chop position (HCYCLE1) are ' \
                                             'the same. Skipping the current image.'

        assert warning[1].message.args[0] == 'The number of images is not equal for chop A and ' \
                                             'chop B.'

    def test_odd_number_images(self) -> None:

        with fits.open(self.fitsfile) as hdulist:
            del hdulist[11]
            hdulist.writeto(self.fitsfile, overwrite=True)

        module = NearReadingModule(name_in='read8',
                                   input_dir=self.test_dir+'near',
                                   chopa_out_tag=self.positions[0],
                                   chopb_out_tag=self.positions[1])

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('read8')

        assert len(warning) == 2

        assert warning[0].message.args[0] == f'FITS file contains odd number of images: ' \
                                             f'{self.fitsfile}'

        assert warning[1].message.args[0] == 'The number of chop cycles (5) is not equal to ' \
                                             'half the number of available HDU images (4).'



"""
Modules for accessing data and attributes in the central database.
"""

import os
import warnings

from abc import ABCMeta, abstractmethod
from typing import Dict, List, Optional, Tuple, Union

import h5py
import numpy as np

from typeguard import typechecked

from pynpoint.util.type_aliases import NonStaticAttribute, StaticAttribute


class DataStorage:
    """
    Instances of DataStorage manage to open and close the Pypeline HDF5 databases. They have an
    internal h5py data bank (self.m_data_bank) which gives direct access to the data if the storage
    is open (self.m_open == True).
    """

    @typechecked
    def __init__(self,
                 location_in: str) -> None:
        """
        Constructor of a DataStorage instance. It needs the location of the HDF5 file (Pypeline
        database) as input. If the file already exists it is opened and extended, if not a new File
        will be created.

        Parameters
        ----------
        location_in : str
            Location (directory + filename) of the HDF5 database.

        Returns
        -------
        NoneType
            None
        """

        assert (os.path.isdir(os.path.split(location_in)[0])), 'Input directory for DataStorage ' \
                                                               'does not exist - input requested:'\
                                                               ' %s.' % location_in

        self._m_location = location_in
        self.m_data_bank = None
        self.m_open = False

    @typechecked
    def open_connection(self) -> None:
        """
        Opens the connection to the HDF5 file by opening an old file or creating a new one.

        Returns
        -------
        NoneType
            None
        """

        if not self.m_open:
            self.m_data_bank = h5py.File(self._m_location, mode='a')
            self.m_open = True

    @typechecked
    def close_connection(self) -> None:
        """
        Closes the connection to the HDF5 file. All entries of the data bank will be stored on the
        hard drive and the memory is cleaned.

        Returns
        -------
        NoneType
            None
        """

        if self.m_open:
            self.m_data_bank.close()
            self.m_data_bank = None
            self.m_open = False


class Port(metaclass=ABCMeta):
    """
    Abstract interface and implementation of common functionality of the InputPort, OutputPort, and
    ConfigPort. Each Port has a internal tag which is its key to a dataset in the DataStorage. If
    for example data is stored under the entry ``im_arr`` in the central data storage only a port
    with the tag (``self._m_tag = im_arr``) can access and change that data. A port knows exactly
    one DataStorage instance, whether it is active or not (``self._m_data_base_active``).
    """

    @abstractmethod
    @typechecked
    def __init__(self,
                 tag: str,
                 data_storage_in: Optional[DataStorage] = None) -> None:
        """
        Abstract constructor of a Port. As input the tag / key is expected which is needed to build
        the connection to the database entry with the same tag / key. It is possible to give the
        Port a DataStorage. If this storage is not given the Pypeline module has to set it or the
        connection needs to be added manually using
        :func:`~pynpoint.core.dataio.Port.set_database_connection`.

        Parameters
        ----------
        tag : str
            Input Tag.
        data_storage_in : pynpoint.core.dataio.DataStorage
            The data storage to which the port is connected.

        Returns
        -------
        NoneType
            None
        """

        assert isinstance(tag, str), 'Port tag needs to be a string.'

        self._m_tag = tag
        self._m_data_storage = data_storage_in
        self._m_data_base_active = False

    @property
    @typechecked
    def tag(self) -> str:
        """
        Getter for the internal tag (no setter).

        Returns
        -------
        str
            Database tag name.
        """

        return self._m_tag

    @typechecked
    def open_port(self) -> None:
        """
        Opens the connection to the :class:`~pynpoint.core.dataio.DataStorage` and activates its
        data bank.

        Returns
        -------
        NoneType
            None
        """

        if not self._m_data_base_active:
            self._m_data_storage.open_connection()
            self._m_data_base_active = True

    @typechecked
    def close_port(self) -> None:
        """
        Closes the connection to the :class:`~pynpoint.core.dataio.DataStorage` and forces it to
        save the data to the hard drive. All data that was accessed using the port is cleaned from
        the memory.

        Returns
        -------
        NoneType
            None
        """

        if self._m_data_base_active:
            self._m_data_storage.close_connection()
            self._m_data_base_active = False

    @typechecked
    def set_database_connection(self,
                                data_base_in: DataStorage) -> None:
        """
        Sets the internal DataStorage instance.

        Parameters
        ----------
        data_base_in: pynpoint.core.dataio.DataStorage
            The input DataStorage.

        Returns
        -------
        NoneType
            None
        """

        self._m_data_storage = data_base_in


class ConfigPort(Port):
    """
    ConfigPort can be used to read the 'config' tag from a (HDF5) database. This tag contains
    the central settings used by PynPoint, as well as the relevant FITS header keywords. You can
    use a ConfigPort instance to access a single attribute of the dataset using get_attribute().
    """

    @typechecked
    def __init__(self,
                 tag: str,
                 data_storage_in: Optional[DataStorage] = None) -> None:
        """
        Constructor of the ConfigPort class which creates the config port instance which can read
        the settings stored in the central database under the tag `config`. An instance of the
        ConfigPort is created in the constructor of PypelineModule such that the attributes in
        the ConfigPort can be accessed from within all type of modules. For example:

        .. code-block:: python

            memory = self._m_config_port.get_attribute('MEMORY')

        Parameters
        ----------
        tag : str
            The tag name of the port. The port can be used to get data from the dataset with the
            key `config`.
        data_storage_in : pynpoint.core.dataio.DataStorage
            The input DataStorage. It is possible to give the constructor of an ConfigPort a
            DataStorage instance which will link the port to that DataStorage. Usually the
            DataStorage is set later by calling
            :func:`~pynpoint.core.dataio.Port.set_database_connection`.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(tag, data_storage_in)

        if tag != 'config':
            raise ValueError('The tag name of the central configuration should be \'config\'.')

    @typechecked
    def _check_status_and_activate(self) -> bool:
        """
        Internal function which checks if the ConfigPort is ready to use and open it.

        Returns
        -------
        bool
            Returns True if the ConfigPort can be used, False if not.
        """

        if self._m_data_storage is None:
            warnings.warn('ConfigPort can not load data unless a database is connected.')
            status = False

        else:
            if not self._m_data_base_active:
                self.open_port()

            status = True

        return status

    @typechecked
    def _check_if_data_exists(self) -> bool:
        """
        Internal function which checks if data exists for the 'config' tag.

        Returns
        -------
        bool
            Returns True if data exists, False if not.
        """

        return 'config' in self._m_data_storage.m_data_bank

    @typechecked
    def _check_error_cases(self) -> bool:
        """'
        Internal function which checks the error cases.
        """

        if not self._check_status_and_activate():
            status = False

        elif self._check_if_data_exists() is False:
            warnings.warn('No data under the tag which is linked by the ConfigPort.')
            status = False

        else:
            status = True

        return status

    @typechecked
    def get_attribute(self,
                      name: str) -> Optional[StaticAttribute]:
        """
        Returns a static attribute which is connected to the dataset of the ConfigPort.

        Parameters
        ----------
        name : str
            The name of the attribute.

        Returns
        -------
        str, float, or int
            The attribute value. Returns None if the attribute does not exist.
        """

        if not self._check_error_cases():
            attr_val = None

        elif name in self._m_data_storage.m_data_bank['config'].attrs:
            attr_val = self._m_data_storage.m_data_bank['config'].attrs[name]

        else:
            warnings.warn(f'The attribute \'{name}\' was not found.')
            attr_val = None

        # Convert numpy types to base types (e.g., np.float64 -> float)
        if isinstance(attr_val, np.generic):
            attr_val = attr_val.item()

        return attr_val


class InputPort(Port):
    """
    InputPorts can be used to read datasets with a specific tag from the HDF5 database. This type
    of port can be used to access:

        * A complete dataset using the get_all() method.
        * A single attribute of the dataset using get_attribute().
        * All attributes of the dataset using get_all_static_attributes() and
          get_all_non_static_attributes().
        * A part of a dataset using slicing. For example:

        .. code-block:: python

            in_port = InputPort('tag')
            data = in_port[0, :, :] # returns the first 2D image of a 3D image stack.

    (More information about how 1D, 2D, and 3D data is organized can be found in the documentation
    of OutputPort (:func:`~pynpoint.core.dataio.OutputPort.append` and
    :func:`~pynpoint.core.dataio.OutputPort.set_all`)

    InputPorts can load two types of attributes which give additional information about
    a dataset the port is linked to:

        * Static attributes: contain global information about a dataset which is not changing
          through a dataset in the database (e.g. the instrument name or pixel scale).
        * Non-static attributes: contain information which changes for different parts of the
          dataset (e.g. the parallactic angles or dithering positions).
    """

    @typechecked
    def __init__(self,
                 tag: str,
                 data_storage_in: Optional[DataStorage] = None) -> None:
        """
        Constructor of InputPort. An input port can read data from the central database under the
        key `tag`. Instances of InputPort should not be created manually inside a PypelineModule
        but should be created with the add_input_port() function.

        Parameters
        ----------
        tag : str
            The tag of the port. The port can be used in order to get data from the dataset with
            the key `tag`.
        data_storage_in : pynpoint.core.dataio.DataStorage
            It is possible to give the constructor of an InputPort a DataStorage instance which
            will link the port to that DataStorage. Usually the DataStorage is set later by calling
            :func:`~pynpoint.core.dataio.Port.set_database_connection`.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(tag, data_storage_in)

        if tag == 'config':
            raise ValueError('The tag name \'config\' is reserved for the central configuration '
                             'of PynPoint.')

        if tag == 'fits_header':
            raise ValueError('The tag name \'fits_header\' is reserved for storage of the FITS '
                             'headers.')

    @typechecked
    def _check_status_and_activate(self) -> bool:
        """
        Internal function which checks if the InputPort is ready to use and open it.

        Returns
        -------
        bool
            Returns True if the InputPort can be used, False if not.
        """

        if self._m_data_storage is None:
            warnings.warn('InputPort can not load data unless a database is connected.')
            status = False

        else:
            status = True

            if not self._m_data_base_active:
                self.open_port()

        return status

    @typechecked
    def _check_if_data_exists(self) -> bool:
        """
        Internal function which checks if data exists for the Port specific tag.

        Returns
        -------
        bool
            Returns True if data exists, False if not.
        """

        return self._m_tag in self._m_data_storage.m_data_bank

    @typechecked
    def _check_error_cases(self) -> bool:

        if not self._check_status_and_activate():
            status = False

        elif self._check_if_data_exists() is False:
            warnings.warn('No data under the tag which is linked by the InputPort.')
            status = False

        else:
            status = True

        return status

    @typechecked
    def __getitem__(self,
                    item: Union[slice, int, tuple]) -> Optional[Union[StaticAttribute,
                                                                      NonStaticAttribute]]:
        """
        Internal function which handles the data access using slicing. See class documentation for a
        example (:class:`~pynpoint.core.dataio.InputPort`). None if the data does not exist.

        Parameters
        ----------
        item : tuple
            Slicing parameter.

        Returns
        -------
        StaticAttribute, NonStaticAttribute, None
            The selected data. Returns None if no data exists under the tag of thePort.
        """

        if not self._check_error_cases():
            data = None

        else:
            data = self._m_data_storage.m_data_bank[self._m_tag][item]

        return data

    @typechecked
    def get_shape(self) -> Optional[Tuple[int, ...]]:
        """
        Returns the shape of the dataset the port is linked to. This can be useful if you need the
        shape without loading the whole data.

        Returns
        -------
        tuple(int, )
            Shape of the dataset. Returns None if the dataset does not exist.
        """

        if not self._check_error_cases():
            data_shape = None

        else:
            self.open_port()
            data_shape = self._m_data_storage.m_data_bank[self._m_tag].shape

        return data_shape

    @typechecked
    def get_ndim(self) -> Optional[int]:
        """
        Returns the number of dimensions of the dataset the port is linked to.

        Returns
        -------
        int
            Number of dimensions of the dataset. Returns None if the dataset does not exist.
        """

        if not self._check_error_cases():
            ndim = None

        else:
            self.open_port()
            ndim = self._m_data_storage.m_data_bank[self._m_tag].ndim

        return ndim

    @typechecked
    def get_all(self) -> Optional[np.ndarray]:
        """
        Returns the whole dataset stored in the data bank under the tag of the Port. Be careful
        using this function for loading large datasets. The data type is inferred from the data
        with numpy.asarray. A 32 bit array will be returned in case the input data is a
        combination of float32 and float64 arrays.

        Returns
        -------
        np.ndarray
            The full dataset. Returns None if the data does not exist.
        """

        if not self._check_error_cases():
            data = None

        else:
            data = np.asarray(self._m_data_storage.m_data_bank[self._m_tag][...])

        return data

    @typechecked
    def get_attribute(self,
                      name: str) -> Optional[Union[StaticAttribute, NonStaticAttribute]]:
        """
        Returns an attribute which is connected to the dataset of the port. The function can return
        static and non-static attributes (static attributes have priority). More information about
        static and non-static attributes can be found in the class documentation of
        :class:`~pynpoint.core.dataio.InputPort`.

        Parameters
        ----------
        name : str
            The name of the attribute.

        Returns
        -------
        StaticAttribute, NonStaticAttribute, None
            The attribute value. Returns None if the attribute does not exist.
        """

        if not self._check_error_cases():
            attr_val = None

        else:
            if name in self._m_data_storage.m_data_bank[self._m_tag].attrs:
                # static attribute
                attr_val = self._m_data_storage.m_data_bank[self._m_tag].attrs[name]

            elif 'header_' + self._m_tag + '/' + name in self._m_data_storage.m_data_bank:
                # non-static attribute
                attribute = 'header_' + self._m_tag + '/' + name
                attr_val = np.asarray(self._m_data_storage.m_data_bank[attribute][...])

            else:
                warnings.warn(f'The attribute \'{name}\' was not found.')
                attr_val = None

        # Convert numpy types to base types (e.g., np.float64 -> float)
        if isinstance(attr_val, np.generic):
            attr_val = attr_val.item()

        return attr_val

    @typechecked
    def get_all_static_attributes(self) -> Optional[Dict[str, StaticAttribute]]:
        """
        Get all static attributes of the dataset which are linked to the Port tag.

        Returns
        -------
        dict, None
            Dictionary of all attributes, as `{attr_name:attr_value}`.
        """

        if not self._check_error_cases():
            attr_dict = None

        else:
            attr_dict = dict(self._m_data_storage.m_data_bank[self._m_tag].attrs)

        return attr_dict

    @typechecked
    def get_all_non_static_attributes(self) -> Optional[List[str]]:
        """
        Returns a list of all non-static attribute keys.  More information about
        static and non-static attributes can be found in the class documentation of
        :class:`~pynpoint.core.dataio.InputPort`.

        Returns
        -------
        list(str, ), None
            List of all existing non-static attribute keys.
        """

        if not self._check_error_cases():
            attr_key = None

        else:
            attr_key = []

            if 'header_' + self._m_tag + '/' in self._m_data_storage.m_data_bank:
                for key in self._m_data_storage.m_data_bank['header_' + self._m_tag + '/']:
                    attr_key.append(key)

            else:
                attr_key = None

        return attr_key


class OutputPort(Port):
    """
    Output ports can be used to save results under a given tag to the HDF5 DataStorage. An instance
    of OutputPort with self.tag=`tag` can store data under the key `tag` by using one of the
    following methods:

        * set_all(...) - replaces and sets the whole dataset
        * append(...) - appends data to the existing data set. For more information see
          function documentation (:func:`~pynpoint.core.dataio.OutputPort.append`).
        * slicing - sets a part of the actual dataset. Example:

        .. code-block:: python

            out_port = OutputPort('Some_tag')
            data = np.ones(200, 200) # 2D image filled with ones
            out_port[0,:,:] = data # Sets the first 2D image of a 3D image stack

        * add_attribute(...) - modifies or creates a attribute of the dataset
        * del_attribute(...) - deletes a attribute
        * del_all_attributes(...) - deletes all attributes
        * append_attribute_data(...) - appends information to non-static attributes. See
          add_attribute() (:func:`~pynpoint.core.dataio.OutputPort.add_attribute`) for more
          information about static and non-static attributes.
        * check_static_attribute(...) - checks if a static attribute exists and if it is equal to a
          given value
        * other functions listed below

    For more information about how data is organized inside the central database have a look at the
    function documentation of the function :func:`~pynpoint.core.dataio.OutputPort.set_all` and
    :func:`~pynpoint.core.dataio.OutputPort.append`.

    Furthermore it is possible to deactivate a OutputPort to stop him saving data.
    """

    @typechecked
    def __init__(self,
                 tag: str,
                 data_storage_in: Optional[DataStorage] = None,
                 activate_init: bool = True) -> None:
        """
        Constructor of the OutputPort class which creates an output port instance which can write
        data to the the central database under the tag `tag`. If you write a PypelineModule you
        should not create instances manually! Use the add_output_port() function instead.

        Parameters
        ----------
        tag : str
            The tag of the port. The port can be used in order to write data to the dataset with
            the key = `tag`.
        data_storage_in : pynpoint.core.dataio.DataStorage
            It is possible to give the constructor of an OutputPort a DataStorage instance which
            will link the port to that DataStorage. Usually the DataStorage is set later by calling
            :func:`~pynpoint.core.dataio.Port.set_database_connection`.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(tag, data_storage_in)

        self.m_activate = activate_init

        if tag == 'config':
            raise ValueError('The tag name \'config\' is reserved for the central configuration '
                             'of PynPoint.')

        if tag == 'fits_header':
            raise ValueError('The tag name \'fits_header\' is reserved for storage of the FITS '
                             'headers.')

    @typechecked
    def _check_status_and_activate(self) -> bool:
        """
        Internal function which checks if the OutputPort is ready to use and open it.

        Returns
        -------
        :return: Returns True if the OutputPort can be used, False if not.
        :rtype: bool
        """

        if not self.m_activate:
            status = False

        elif self._m_data_storage is None:
            warnings.warn('OutputPort can not store data unless a database is connected.')
            status = False

        else:
            if not self._m_data_base_active:
                self.open_port()

            status = True

        return status

    @typechecked
    def _init_dataset(self,
                      first_data: Union[np.ndarray, list],
                      tag: str,
                      data_dim: Optional[int] = None) -> None:
        """
        Internal function which is used to initialize a dataset in the HDF5 database.

        Parameters
        ----------
        first_data : np.ndarray, list
            The initial data.
        tag : str
            Database tag.
        data_dim : int, None
            Number of dimensions. The dimensions of ``first_data`` is used if set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        @typechecked
        def _ndim_check(data_dim: int,
                        first_dim: int) -> None:

            if first_dim > 5 or first_dim < 1:
                raise ValueError('Output port can only save numpy arrays from 1D to 5D. Use Port '
                                 'attributes to save as int, float, or string.')

            if data_dim > 5 or data_dim < 1:
                raise ValueError('The data dimensions should be 1D, 2D, 3D, 4D, or 5D.')

            if data_dim < first_dim:
                raise ValueError('The dimensions of the data should be equal to or larger than the '
                                 'dimensions of the input data.')

            if data_dim == 3 and first_dim == 1:
                raise ValueError('Cannot initialize 1D data in 3D data container.')

        first_data = np.asarray(first_data)

        if data_dim is None:
            data_dim = first_data.ndim

        _ndim_check(data_dim, first_data.ndim)

        if data_dim == first_data.ndim:
            if first_data.ndim == 1:  # 1D
                data_shape = (None, )

            elif first_data.ndim == 2:  # 2D
                data_shape = (None, first_data.shape[1])

            elif first_data.ndim == 3:  # 3D
                data_shape = (None, first_data.shape[1], first_data.shape[2])

            elif first_data.ndim == 4:  # 4D
                data_shape = (first_data.shape[0], None, first_data.shape[2], first_data.shape[3])

            elif first_data.ndim == 5:  # 5D
                data_shape = (first_data.shape[0], first_data.shape[1], first_data.shape[2],
                              first_data.shape[3], first_data.shape[4])

        else:
            if data_dim == 2:  # 1D -> 2D
                data_shape = (None, first_data.shape[0])
                first_data = first_data[np.newaxis, :]

            elif data_dim == 3:  # 2D -> 3D
                data_shape = (None, first_data.shape[0], first_data.shape[1])
                first_data = first_data[np.newaxis, :, :]

            elif data_dim == 4:  # 3D -> 4D
                data_shape = (first_data.shape[0], None, first_data.shape[1], first_data.shape[2])
                first_data = first_data[:, np.newaxis, :, :]

        if first_data.size == 0:
            warnings.warn(f'The new dataset that is stored under the tag name \'{tag}\' is empty.')

        else:
            if isinstance(first_data[0], str):
                first_data = np.array(first_data, dtype='|S')

        self._m_data_storage.m_data_bank.create_dataset(tag,
                                                        data=first_data,
                                                        maxshape=data_shape)

    @typechecked
    def _set_all_key(self,
                     tag: str,
                     data: np.ndarray,
                     data_dim: Optional[int] = None,
                     keep_attributes: bool = False) -> None:
        """
        Internal function which sets the values of a dataset under the *tag* name in the database.
        If old data exists it will be overwritten. This function is used by
        :func:`~pynpoint.core.dataio.OutputPort.set_all` and for setting non-static attributes.

        Parameters
        ----------
        tag : str
            Database tag of the data that will be modified.
        data : np.ndarray
            The data that will be stored and replace any old data.
        data_dim : int
            Number of dimension of the data.
        keep_attributes : bool
            Keep all static attributes of the dataset if set to True. Non-static attributes will be
            kept anyway so not needed for setting non-static attributes.

        Returns
        -------
        NoneType
            None
        """

        tmp_attributes = {}

        # check if database entry is new...
        if tag in self._m_data_storage.m_data_bank:
            # NO -> database entry exists
            if keep_attributes:
                # we have to copy all attributes since deepcopy is not supported
                for key, value in self._m_data_storage.m_data_bank[tag].attrs.items():
                    tmp_attributes[key] = value

            # remove database entry
            del self._m_data_storage.m_data_bank[tag]

        # make new database entry
        self._init_dataset(data, tag, data_dim=data_dim)

        if keep_attributes:
            for key, value in tmp_attributes.items():
                self._m_data_storage.m_data_bank[tag].attrs[key] = value

    @typechecked
    def _append_key(self,
                    tag: str,
                    data: Union[np.ndarray, list],
                    data_dim: Optional[int] = None,
                    force: bool = False) -> None:
        """
        Internal function for appending data to a dataset or appending non-static attributes.
        See :func:`~pynpoint.core.dataio.OutputPort.append` for more information.

        Parameters
        ----------
        tag : str
            Database tag where the data will be stored.
        data : np.ndarray
            The data that will be appended.
        data_dim : int
            Number of dimension of the data.
        force : bool
            The existing data will be overwritten if shape or type does not match.

        Returns
        -------
        NoneType
            None
        """

        # check if database entry is new...
        if tag not in self._m_data_storage.m_data_bank:
            # YES -> database entry is new
            self._init_dataset(data, tag, data_dim=data_dim)
            return None

        # NO -> database entry exists
        # check if the existing data has the same dim and datatype
        tmp_shape = self._m_data_storage.m_data_bank[tag].shape
        tmp_dim = len(tmp_shape)

        if data_dim is None:
            data_dim = tmp_dim

        # convert input data to numpy array
        data = np.asarray(data)

        # if the dimension offset is 1 add that dimension (e.g. save 2D image in 3D image stack)
        if data.ndim + 1 == data_dim:

            if data_dim == 2:
                data = data[np.newaxis, :]

            elif data_dim == 3:
                data = data[np.newaxis, :, :]

            elif data_dim == 4:
                data = data[:, np.newaxis, :, :]

        @typechecked
        def _type_check() -> bool:
            check_result = False

            if tmp_dim == data.ndim:

                if tmp_dim == 1:
                    check_result = True

                elif tmp_dim == 2:
                    check_result = tmp_shape[1] == data.shape[1]

                elif tmp_dim == 3:
                    # check if the spatial shape is the same
                    check_result = (tmp_shape[1] == data.shape[1]) and \
                                   (tmp_shape[2] == data.shape[2])

                elif tmp_dim == 4:
                    # check if the spectral and spatial shape is the same
                    check_result = (tmp_shape[0] == data.shape[0]) and \
                                   (tmp_shape[2] == data.shape[2]) and \
                                   (tmp_shape[3] == data.shape[3])

            return check_result

        if _type_check():
            # YES -> dim and type match
            # we always append in axis one independent of the dimension
            # 1D case

            if data.size == 0:
                warnings.warn(f'The dataset that is appended under the tag name \'{tag}\' '
                              f'is empty.')

            else:
                if isinstance(data[0], str):
                    data = np.array(data, dtype='|S')

            if data.ndim == 4:
                # IFS data: (n_wavelength, n_dit, y_pos, x_pos)
                self._m_data_storage.m_data_bank[tag].resize(tmp_shape[1] + data.shape[1], axis=1)
                self._m_data_storage.m_data_bank[tag][:, tmp_shape[1]:, :, :] = data

            else:
                # Other data: n_dit is the first dimension
                self._m_data_storage.m_data_bank[tag].resize(tmp_shape[0] + data.shape[0], axis=0)
                self._m_data_storage.m_data_bank[tag][tmp_shape[0]:, ] = data

            return None

        # NO -> shape or type is different
        # Check force
        if force:
            # YES -> Force is true
            self._set_all_key(tag, data=data)
            return None

        # NO -> Error message
        raise ValueError(f'The port tag \'{self._m_tag}\' is already used with a different data '
                         f'type. The \'force\' parameter can be used to replace the tag.')

    @typechecked
    def __setitem__(self,
                    key: Union[slice, int, tuple],
                    value: Union[np.ndarray, int]) -> None:
        """
        Internal function needed to change data using slicing. See class documentation for an
        example (:class:`~pynpoint.core.dataio.OutputPort`).

        Parameters
        ----------
        key : slice
            Index slice to be changed.
        value : np.ndarray
            New data.

        Returns
        -------
        NoneType
            None
        """

        if self._check_status_and_activate():
            self._m_data_storage.m_data_bank[self._m_tag][key] = value

    @typechecked
    def del_all_data(self) -> None:
        """
        Delete all data belonging to the database tag.
        """

        if self._check_status_and_activate():

            if self._m_tag in self._m_data_storage.m_data_bank:
                del self._m_data_storage.m_data_bank[self._m_tag]

    @typechecked
    def set_all(self,
                data: Union[np.ndarray, list],
                data_dim: Optional[int] = None,
                keep_attributes: bool = False) -> None:
        """
        Set the data in the database by replacing all old values with the values of the input data.
        If no old values exists the data is just stored. Since it is not possible to change the
        number of dimensions of a data set later in the processing history one can choose a
        dimension different to the input data. The following cases are implemented:

            * (#dimension of the first input data#, #desired data_dim#)
            * (1, 1) 1D input or single value will be stored as list in HDF5
            * (1, 2) 1D input, but 2D array stored inside (i.e. a list of lists with a fixed size).
            * (2, 2) 2D input (single image) and 2D array stored inside (i.e. a list of lists with a
              fixed size).
            * (2, 3) 2D input (single image) but 3D array stored inside (i.e. a stack of images with
              a fixed size).
            * (3, 3) 3D input and 3D array stored inside (i.e. a stack of images with a fixed size).

        For 2D and 3D data the first dimension always represents the list / stack (variable size)
        while the second (or third) dimension has a fixed size. After creation it is possible to
        extend a data set using :func:`~pynpoint.core.dataio.OutputPort.append` along the first
        dimension.

        **Example 1:**

        Input 2D array with size (200, 200). Desired dimension 3D. The result is a 3D dataset with
        the dimension (1, 200, 200). It is possible to append other images with the size (200, 200)
        or other stacks of images with the size (:, 200, 200).

        **Example 2:**

        Input 2D array with size (200, 200). Desired dimension 2D. The result is a 2D dataset with
        the dimension (200, 200). It is possible to append other list with the length 200
        or other stacks of lines with the size (:, 200). However it is not possible to append other
        2D images along a third dimension.

        Parameters
        ----------
        data : np.ndarray
            The data to be saved.
        data_dim : int
            Number of data dimensions. The dimension of the *first_data* is used if set to None.
        keep_attributes : bool
            All attributes of the old dataset will remain the same if set to True.

        Returns
        -------
        NoneType
            None
        """

        data = np.asarray(data)

        if self._check_status_and_activate():

            self._set_all_key(tag=self._m_tag,
                              data=data,
                              data_dim=data_dim,
                              keep_attributes=keep_attributes)

    @typechecked
    def append(self,
               data: Union[np.ndarray, list],
               data_dim: Optional[int] = None,
               force: bool = False) -> None:
        """
        Appends data to an existing dataset along the first dimension. If no data exists for the
        :class:`~pynpoint.core.dataio.OutputPort`, then a new data set is created. For more
        information about how the dimensions are organized, see the documentation of
        :func:`~pynpoint.core.dataio.OutputPort.set_all`. Note it is not possible to append data
        with a different shape or data type to an existing dataset.

        **Example:** An internal data set is 3D (storing a stack of 2D images) with shape of
        ``(233, 300, 300)``, that is, it contains 233 images with a resolution of 300 by 300
        pixels. Thus it is only possible to extend along the first dimension by appending new
        images with a shape of ``(300, 300)`` or by appending a stack of images with a shape of
        ``(:, 300, 300)``.

        It is possible to force the function to overwrite existing data set if the shape or type of
        the input data do not match the existing data.

        Parameters
        ----------
        data : np.ndarray
            The data that will be appended.
        data_dim : int
            Number of data dimensions used if a new data set is created. The dimension of the
            ``data`` is used if set to None.
        force : bool
            The existing data will be overwritten if the shape or type does not match.

        Returns
        -------
        NoneType
            None
        """

        if self._check_status_and_activate():

            self._append_key(self._m_tag,
                             data=data,
                             data_dim=data_dim,
                             force=force)

    @typechecked
    def activate(self) -> None:
        """
        Activates the port. A non activated port will not save data.

        Returns
        -------
        NoneType
            None
        """

        self.m_activate = True

    @typechecked
    def deactivate(self) -> None:
        """
        Deactivates the port. A non activated port will not save data.

        Returns
        -------
        NoneType
            None
        """

        self.m_activate = False

    @typechecked
    def add_attribute(self,
                      name: str,
                      value: Union[StaticAttribute, NonStaticAttribute],
                      static: bool = True) -> None:
        """
        Adds an attribute to the dataset of the Port with the attribute name = `name` and the
        value = `value`. If the attribute already exists it will be overwritten. Two different
        types of attributes are supported:

            1. **static attributes**:
               Contain a single value or name (e.g. The name of the used Instrument).
            2. **non-static attributes**:
               Contain a dataset which is connected to the actual data set (e.g. Instrument
               temperature). It is possible to append additional information to non-static
               attributes later (:func:`~pynpoint.core.dataio.OutputPort.append_attribute_data`).
               This is not supported by static attributes.

        Static and non-static attributes are stored in a different way using the HDF5 file format.
        Static attributes will be direct attributes while non-static attributes are stored in a
        group with the name *header_* + name of the dataset.

        Parameters
        ----------
        name : str
            Name of the attribute.
        value : StaticAttribute, NonStaticAttribute
            Value of the attribute.
        static : bool
            Indicate if the attribute is static (True) or non-static (False).

        Returns
        -------
        NoneType
            None
        """

        if self._check_status_and_activate():

            if self._m_tag not in self._m_data_storage.m_data_bank:
                warnings.warn(f'Can not store the attribute \'{name}\' because the dataset '
                              f'\'{self._m_tag}\' does not exist.')

            else:
                if static:
                    self._m_data_storage.m_data_bank[self._m_tag].attrs[name] = value

                else:
                    self._set_all_key(tag=('header_' + self._m_tag + '/' + name),
                                      data=np.asarray(value))

    @typechecked
    def append_attribute_data(self,
                              name: str,
                              value: Union[StaticAttribute, NonStaticAttribute]) -> None:
        """
        Function which appends data (either a single value or an array) to non-static attributes.

        Parameters
        ----------
        name : str
            Name of the attribute.
        value : StaticAttribute, NonStaticAttribute
            Value which will be appended to the attribute dataset.

        Returns
        -------
        NoneType
            None
        """

        if self._check_status_and_activate():

            self._append_key(tag=('header_' + self._m_tag + '/' + name),
                             data=np.asarray([value, ]))

    @typechecked
    def copy_attributes(self,
                        input_port: InputPort) -> None:
        """
        Copies all static and non-static attributes from a given InputPort. Attributes which already
        exist will be overwritten. Non-static attributes will be linked not copied. If the InputPort
        tag = OutputPort tag (self.tag) nothing will be changed. Use this function in all modules
        to keep the header information.

        Parameters
        ----------
        input_port : pynpoint.core.dataio.InputPort
            The InputPort with the header information.

        Returns
        -------
        NoneType
            None
        """

        if self._check_status_and_activate() and input_port.tag != self._m_tag:

            # link non-static attributes
            if 'header_' + input_port.tag + '/' in self._m_data_storage.m_data_bank:

                for attr_name, attr_data in self._m_data_storage\
                        .m_data_bank['header_' + input_port.tag + '/'].items():

                    database_name = 'header_'+self._m_tag+'/'+attr_name

                    # overwrite existing header information in the database
                    if database_name in self._m_data_storage.m_data_bank:
                        del self._m_data_storage.m_data_bank[database_name]

                    self._m_data_storage.m_data_bank[database_name] = attr_data

            # copy static attributes
            attributes = input_port.get_all_static_attributes()
            for attr_name, attr_val in attributes.items():
                self.add_attribute(attr_name, attr_val)

            self._m_data_storage.m_data_bank.flush()

    @typechecked
    def del_attribute(self,
                      name: str) -> None:
        """
        Deletes the attribute of the dataset with the given name. Finds and removes static and
        non-static attributes.

        Parameters
        ----------
        name : str
            Name of the attribute.

        Returns
        -------
        NoneType
            None
        """

        if self._check_status_and_activate():

            # check if attribute is static
            if name in self._m_data_storage.m_data_bank[self._m_tag].attrs:
                del self._m_data_storage.m_data_bank[self._m_tag].attrs[name]

            elif 'header_'+self._m_tag+'/'+name in self._m_data_storage.m_data_bank:
                # remove non-static attribute
                del self._m_data_storage.m_data_bank[('header_' + self._m_tag + '/' + name)]

            else:
                warnings.warn(f'Attribute \'{name}\' does not exist and could not be deleted.')

    @typechecked
    def del_all_attributes(self) -> None:
        """
        Deletes all static and non-static attributes of the dataset.

        Returns
        -------
        NoneType
            None
        """

        if self._check_status_and_activate():

            # static attributes
            if self._m_tag in self._m_data_storage.m_data_bank:
                self._m_data_storage.m_data_bank[self._m_tag].attrs.clear()

            # non-static attributes
            if 'header_' + self._m_tag + '/' in self._m_data_storage.m_data_bank:
                del self._m_data_storage.m_data_bank[('header_' + self._m_tag + '/')]

    @typechecked
    def check_static_attribute(self,
                               name: str,
                               comparison_value: StaticAttribute) -> Optional[int]:
        """
        Checks if a static attribute exists and if it is equal to a comparison value.

        Parameters
        ----------
        name : str
            Name of the static attribute.
        comparison_value : StaticAttribute
            Comparison value.

        Returns
        -------
        int, None
            Status: 1 if the static attribute does not exist, 0 if the static attribute exists
            and is equal, and -1 if the static attribute exists but is not equal.
        """

        if not self._check_status_and_activate():
            return None

        if name in self._m_data_storage.m_data_bank[self._m_tag].attrs:
            if self._m_data_storage.m_data_bank[self._m_tag].attrs[name] == comparison_value:
                return 0

            return -1

        return 1

    @typechecked
    def check_non_static_attribute(self,
                                   name: str,
                                   comparison_value: NonStaticAttribute) -> Optional[int]:
        """
        Checks if a non-static attribute exists and if it is equal to a comparison value.

        Parameters
        ----------
        name : str
            Name of the non-static attribute.
        comparison_value : NonStaticAttribute
            Comparison values

        Returns
        -------
        int, None
            Status: 1 if the non-static attribute does not exist, 0 if the non-static attribute
            exists and is equal, and -1 if the non-static attribute exists but is not equal.
        """

        if not self._check_status_and_activate():
            return None

        group = 'header_' + self._m_tag + '/'

        if group in self._m_data_storage.m_data_bank:
            if name in self._m_data_storage.m_data_bank[group]:
                if np.array_equal(self._m_data_storage.m_data_bank[group+name][:],
                                  comparison_value):
                    return 0

                return -1

            return 1

        return 1

    @typechecked
    def add_history(self,
                    module: str,
                    history: str) -> None:
        """
        Adds an attribute with history information about the pipeline module.

        Parameters
        ----------
        module : str
            Name of the pipeline module which was executed.
        history : str
            History information.

        Returns
        -------
        NoneType
            None
        """

        self.add_attribute('History: ' + module, history)

    @typechecked
    def flush(self) -> None:
        """
        Forces the :class:`~pynpoint.core.dataio.DataStorage` to save all data from the memory to
        the hard drive without closing the :class:`~pynpoint.core.dataio.OutputPort`.

        Returns
        -------
        NoneType
            None
        """

        self._m_data_storage.m_data_bank.flush()



import os

import numpy as np
import pytest

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.background import LineSubtractionModule, \
                                           MeanBackgroundSubtractionModule, \
                                           NoddingBackgroundModule, \
                                           SimpleBackgroundSubtractionModule
from pynpoint.processing.pcabackground import DitheringBackgroundModule
from pynpoint.processing.stacksubset import StackCubesModule
from pynpoint.util.tests import create_config, create_dither_data, create_star_data, \
                                remove_test_data


class TestBackground:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_dither_data(self.test_dir+'dither')
        create_star_data(self.test_dir+'science')
        create_star_data(self.test_dir+'sky')
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['dither', 'science', 'sky'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read1',
                                   image_tag='dither',
                                   input_dir=self.test_dir+'dither')

        self.pipeline.add_module(module)
        self.pipeline.run_module('read1')

        data = self.pipeline.get_data('dither')
        assert np.sum(data) == pytest.approx(211.20534661914408, rel=self.limit, abs=0.)
        assert data.shape == (20, 21, 21)

        module = FitsReadingModule(name_in='read2',
                                   image_tag='science',
                                   input_dir=self.test_dir+'science')

        self.pipeline.add_module(module)
        self.pipeline.run_module('read2')

        data = self.pipeline.get_data('science')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

        module = FitsReadingModule(name_in='read3',
                                   image_tag='sky',
                                   input_dir=self.test_dir+'sky')

        self.pipeline.add_module(module)
        self.pipeline.run_module('read3')

        data = self.pipeline.get_data('sky')
        assert np.sum(data) == pytest.approx(105.54278879805277, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_simple_background(self) -> None:

        module = SimpleBackgroundSubtractionModule(shift=5,
                                                   name_in='simple',
                                                   image_in_tag='dither',
                                                   image_out_tag='simple')

        self.pipeline.add_module(module)
        self.pipeline.run_module('simple')

        data = self.pipeline.get_data('simple')
        # TODO Unclear why the absolute value is required
        # Local the sum is 3.55 and on Github -3.55
        assert np.abs(np.sum(data)) == pytest.approx(3.552713678800501e-15, rel=self.limit, abs=0.)
        assert data.shape == (20, 21, 21)

    def test_mean_background_shift(self) -> None:

        module = MeanBackgroundSubtractionModule(shift=5,
                                                 cubes=1,
                                                 name_in='mean2',
                                                 image_in_tag='dither',
                                                 image_out_tag='mean2')

        self.pipeline.add_module(module)
        self.pipeline.run_module('mean2')

        data = self.pipeline.get_data('mean2')
        assert np.sum(data) == pytest.approx(2.473864361018551, rel=self.limit, abs=0.)
        assert data.shape == (20, 21, 21)

    def test_mean_background_nframes(self) -> None:

        module = MeanBackgroundSubtractionModule(shift=None,
                                                 cubes=1,
                                                 name_in='mean1',
                                                 image_in_tag='dither',
                                                 image_out_tag='mean1')

        self.pipeline.add_module(module)
        self.pipeline.run_module('mean1')

        data = self.pipeline.get_data('mean1')
        assert np.sum(data) == pytest.approx(2.473864361018551, rel=self.limit, abs=0.)
        assert data.shape == (20, 21, 21)

    def test_dithering_attributes(self) -> None:

        module = DitheringBackgroundModule(name_in='pca_dither1',
                                           image_in_tag='dither',
                                           image_out_tag='pca_dither1',
                                           center=None,
                                           cubes=None,
                                           size=0.2,
                                           gaussian=0.05,
                                           subframe=0.1,
                                           pca_number=1,
                                           mask_star=0.05)

        self.pipeline.add_module(module)
        self.pipeline.run_module('pca_dither1')

        data = self.pipeline.get_data('dither_dither_crop1')
        assert np.sum(data) == pytest.approx(54.62410860562912, rel=self.limit, abs=0.)
        assert data.shape == (20, 9, 9)

        data = self.pipeline.get_data('dither_dither_star1')
        assert np.sum(data) == pytest.approx(54.873885838788595, rel=self.limit, abs=0.)
        assert data.shape == (5, 9, 9)

        data = self.pipeline.get_data('dither_dither_mean1')
        assert np.sum(data) == pytest.approx(54.204960755115245, rel=self.limit, abs=0.)
        assert data.shape == (5, 9, 9)

        data = self.pipeline.get_data('dither_dither_background1')
        assert np.sum(data) == pytest.approx(-0.24977723315947564, rel=self.limit, abs=0.)
        assert data.shape == (15, 9, 9)

        data = self.pipeline.get_data('dither_dither_pca_fit1')
        assert np.sum(data) == pytest.approx(-0.6816458444287745, rel=1e-5, abs=0.)
        assert data.shape == (5, 9, 9)

        data = self.pipeline.get_data('dither_dither_pca_res1')
        assert np.sum(data) == pytest.approx(55.63879076093719, rel=1e-6, abs=0.)
        assert data.shape == (5, 9, 9)

        data = self.pipeline.get_data('dither_dither_pca_mask1')
        assert np.sum(data) == pytest.approx(360.0, rel=self.limit, abs=0.)
        assert data.shape == (5, 9, 9)

        data = self.pipeline.get_data('pca_dither1')
        assert np.sum(data) == pytest.approx(208.24417329569593, rel=1e-6, abs=0.)
        assert data.shape == (20, 9, 9)

        attr = self.pipeline.get_attribute('dither_dither_pca_res1', 'STAR_POSITION', static=False)
        assert np.sum(attr) == pytest.approx(40., rel=self.limit, abs=0.)
        assert attr.shape == (5, 2)

    def test_dithering_center(self) -> None:

        module = DitheringBackgroundModule(name_in='pca_dither2',
                                           image_in_tag='dither',
                                           image_out_tag='pca_dither2',
                                           center=[(5, 5), (5, 15), (15, 15), (15, 5)],
                                           cubes=1,
                                           size=0.2,
                                           gaussian=0.05,
                                           subframe=None,
                                           pca_number=1,
                                           mask_star=0.05)

        self.pipeline.add_module(module)
        self.pipeline.run_module('pca_dither2')

        data = self.pipeline.get_data('pca_dither2')
        assert np.sum(data) == pytest.approx(208.24417332523367, rel=1e-6, abs=0.)
        assert data.shape == (20, 9, 9)

    def test_nodding_background(self) -> None:

        module = StackCubesModule(name_in='mean',
                                  image_in_tag='sky',
                                  image_out_tag='mean',
                                  combine='mean')

        self.pipeline.add_module(module)
        self.pipeline.run_module('mean')

        data = self.pipeline.get_data('mean')
        assert np.sum(data) == pytest.approx(21.108557759610548, rel=self.limit, abs=0.)
        assert data.shape == (2, 11, 11)

        attr = self.pipeline.get_attribute('mean', 'INDEX', static=False)
        assert np.sum(attr) == pytest.approx(1, rel=self.limit, abs=0.)
        assert attr.shape == (2, )

        attr = self.pipeline.get_attribute('mean', 'NFRAMES', static=False)
        assert np.sum(attr) == pytest.approx(2, rel=self.limit, abs=0.)
        assert attr.shape == (2, )

        module = NoddingBackgroundModule(name_in='nodding',
                                         sky_in_tag='mean',
                                         science_in_tag='science',
                                         image_out_tag='nodding',
                                         mode='both')

        self.pipeline.add_module(module)
        self.pipeline.run_module('nodding')

        data = self.pipeline.get_data('nodding')
        assert np.sum(data) == pytest.approx(1.793466459074705, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_line_background_mean(self) -> None:

        module = LineSubtractionModule(name_in='line1',
                                       image_in_tag='science',
                                       image_out_tag='science_line1',
                                       combine='mean',
                                       mask=0.1)

        self.pipeline.add_module(module)
        self.pipeline.run_module('line1')

        data = self.pipeline.get_data('science_line1')
        assert np.sum(data) == pytest.approx(104.55443019231085, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_line_background_median(self) -> None:

        module = LineSubtractionModule(name_in='line2',
                                       image_in_tag='science',
                                       image_out_tag='science_line2',
                                       combine='median',
                                       mask=0.1)

        self.pipeline.add_module(module)
        self.pipeline.run_module('line2')

        data = self.pipeline.get_data('science_line2')
        assert np.sum(data) == pytest.approx(106.09825573198366, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)



"""
Functions for post-processing.
"""

from typing import Union, Optional, Tuple

import numpy as np

from typeguard import typechecked
from sklearn.decomposition import PCA

from pynpoint.util.psf import pca_psf_subtraction
from pynpoint.util.sdi import sdi_scaling


@typechecked
def postprocessor(images: np.ndarray,
                  angles: np.ndarray,
                  scales: Optional[np.ndarray],
                  pca_number: Union[int, Tuple[Union[int, np.int32, np.int64], Union[int, np.int32, np.int64]]],
                  pca_sklearn: Optional[PCA] = None,
                  im_shape: Union[None, tuple] = None,
                  indices: Optional[np.ndarray] = None,
                  mask: Optional[np.ndarray] = None,
                  processing_type: str = 'ADI'):

    """
    Function to apply different kind of post processings. It is equivalent to
    :func:`~pynpoint.util.psf.pca_psf_subtraction` if ``processing_type='ADI'` and
    ``mask=None``.

    Parameters
    ----------
    images : np.array
        Input images which should be reduced.
    angles : np.ndarray
        Derotation angles (deg).
    scales : np.array
        Scaling factors
    pca_number : tuple(int, int)
        Number of principal components used for the PSF subtraction.
    pca_sklearn : sklearn.decomposition.pca.PCA, None
        PCA object with the basis if not set to None.
    im_shape : tuple(int, int, int), None
        Original shape of the stack with images. Required if ``pca_sklearn`` is not set to None.
    indices : np.ndarray, None
        Non-masked image indices. All pixels are used if set to None.
    mask : np.ndarray
        Mask (2D).
    processing_type : str
        Post-processing type:
            - ADI: Angular differential imaging.
            - SDI: Spectral differential imaging.
            - SDI+ADI: Spectral and angular differential imaging.
            - ADI+SDI: Angular and spectral differential imaging.

    Returns
    -------
    np.ndarray
        Residuals of the PSF subtraction.
    np.ndarray
        Derotated residuals of the PSF subtraction.
    """

    if not isinstance(pca_number, tuple):
        pca_number = (pca_number, -1)

    if mask is None:
        mask = 1.

    res_raw = np.zeros(images.shape)
    res_rot = np.zeros(images.shape)

    if processing_type == 'ADI':
        if images.ndim == 2:
            res_raw, res_rot = pca_psf_subtraction(images=images*mask,
                                                   angles=angles,
                                                   scales=None,
                                                   pca_number=pca_number[0],
                                                   pca_sklearn=pca_sklearn,
                                                   im_shape=im_shape,
                                                   indices=indices)

        elif images.ndim == 4:
            for i in range(images.shape[0]):
                res_raw[i, ], res_rot[i, ] = pca_psf_subtraction(images=images[i, ]*mask,
                                                                 angles=angles,
                                                                 scales=None,
                                                                 pca_number=pca_number[0],
                                                                 pca_sklearn=pca_sklearn,
                                                                 im_shape=im_shape,
                                                                 indices=indices)

    elif processing_type == 'SDI':
        for i in range(images.shape[1]):
            im_scaled = sdi_scaling(images[:, i, :, :], scales)

            res_raw[:, i], res_rot[:, i] = pca_psf_subtraction(images=im_scaled*mask,
                                                               angles=np.full(scales.size,
                                                                              angles[i]),
                                                               scales=scales,
                                                               pca_number=pca_number[0],
                                                               pca_sklearn=pca_sklearn,
                                                               im_shape=im_shape,
                                                               indices=indices)

    elif processing_type == 'SDI+ADI':
        # SDI
        res_raw_int = np.zeros(res_raw.shape)

        for i in range(images.shape[1]):
            im_scaled = sdi_scaling(images[:, i], scales)

            res_raw_int[:, i], _ = pca_psf_subtraction(images=im_scaled*mask,
                                                       angles=None,
                                                       scales=scales,
                                                       pca_number=pca_number[0],
                                                       pca_sklearn=pca_sklearn,
                                                       im_shape=im_shape,
                                                       indices=indices)

        # ADI
        for i in range(images.shape[0]):
            res_raw[i], res_rot[i] = pca_psf_subtraction(images=res_raw_int[i]*mask,
                                                         angles=angles,
                                                         scales=None,
                                                         pca_number=pca_number[1],
                                                         pca_sklearn=pca_sklearn,
                                                         im_shape=im_shape,
                                                         indices=indices)

    elif processing_type == 'ADI+SDI':
        # ADI
        res_raw_int = np.zeros(res_raw.shape)

        for i in range(images.shape[0]):
            res_raw_int[i], _ = pca_psf_subtraction(images=images[i, ]*mask,
                                                    angles=None,
                                                    scales=None,
                                                    pca_number=pca_number[0],
                                                    pca_sklearn=pca_sklearn,
                                                    im_shape=im_shape,
                                                    indices=indices)

        # SDI
        for i in range(images.shape[1]):
            im_scaled = sdi_scaling(res_raw_int[:, i], scales)

            res_raw[:, i], res_rot[:, i] = pca_psf_subtraction(images=im_scaled*mask,
                                                               angles=np.full(scales.size,
                                                                              angles[i]),
                                                               scales=scales,
                                                               pca_number=pca_number[1],
                                                               pca_sklearn=pca_sklearn,
                                                               im_shape=im_shape,
                                                               indices=indices)

    elif processing_type == 'CODI':
        # flatten images from 4D to 3D
        ims = images.shape
        im_scaled_flat = np.zeros((ims[0]*ims[1], ims[2], ims[3]))
        scales_flat = np.zeros((ims[0]*ims[1]))
        angles_flat = np.zeros((ims[0]*ims[1]))
        for i in range(ims[1]):
            im_scaled_flat[i*ims[0]:(i+1)*ims[0]] = sdi_scaling(images[:, i], scales)
            scales_flat[i*ims[0]:(i+1)*ims[0]] = scales
            angles_flat[i*ims[0]:(i+1)*ims[0]] = angles[i]

        # codi
        res_raw_flat, res_rot_flat = pca_psf_subtraction(images=im_scaled_flat*mask,
                                                         angles=angles_flat,
                                                         scales=scales_flat,
                                                         pca_number=pca_number[0],
                                                         pca_sklearn=pca_sklearn,
                                                         im_shape=im_shape,
                                                         indices=indices)

        # inflate images from 3D to 4D
        for i in range(ims[1]):
            res_raw[:, i] = res_raw_flat[i*ims[0]:(i+1)*ims[0]]
            res_rot[:, i] = res_rot_flat[i*ims[0]:(i+1)*ims[0]]

    return res_raw, res_rot



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.hdf5reading import Hdf5ReadingModule
from pynpoint.readwrite.hdf5writing import Hdf5WritingModule
from pynpoint.util.tests import create_config, create_random, remove_test_data


class TestHdf5Writing:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_random(self.test_dir)
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, files=['test.hdf5'])

    def test_hdf5_writing(self) -> None:

        module = Hdf5WritingModule(name_in='write1',
                                   file_name='test.hdf5',
                                   output_dir=None,
                                   tag_dictionary={'images': 'data1'},
                                   keep_attributes=True,
                                   overwrite=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('write1')

    def test_no_data_tag(self) -> None:

        module = Hdf5WritingModule(name_in='write2',
                                   file_name='test.hdf5',
                                   output_dir=None,
                                   tag_dictionary={'empty': 'empty'},
                                   keep_attributes=True,
                                   overwrite=False)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('write2')

        assert len(warning) == 1
        assert warning[0].message.args[0] == 'No data under the tag which is linked by the ' \
                                             'InputPort.'

    def test_overwrite_false(self) -> None:

        module = Hdf5WritingModule(name_in='write3',
                                   file_name='test.hdf5',
                                   output_dir=None,
                                   tag_dictionary={'images': 'data2'},
                                   keep_attributes=True,
                                   overwrite=False)

        self.pipeline.add_module(module)
        self.pipeline.run_module('write3')

    def test_dictionary_none(self) -> None:

        module = Hdf5WritingModule(name_in='write4',
                                   file_name='test.hdf5',
                                   output_dir=None,
                                   tag_dictionary=None,
                                   keep_attributes=True,
                                   overwrite=False)

        self.pipeline.add_module(module)
        self.pipeline.run_module('write4')

    def test_hdf5_reading(self) -> None:

        module = Hdf5ReadingModule(name_in='read',
                                   input_filename='test.hdf5',
                                   input_dir=self.test_dir,
                                   tag_dictionary={'data1': 'data1', 'data2': 'data2'})

        self.pipeline.add_module(module)
        self.pipeline.run_module('read')

        data1 = self.pipeline.get_data('data1')
        data2 = self.pipeline.get_data('data2')
        data3 = self.pipeline.get_data('images')
        assert data1 == pytest.approx(data2, rel=self.limit, abs=0.)
        assert data2 == pytest.approx(data3, rel=self.limit, abs=0.)

        attribute1 = self.pipeline.get_attribute('images', 'PARANG', static=False)
        attribute2 = self.pipeline.get_attribute('data1', 'PARANG', static=False)
        attribute3 = self.pipeline.get_attribute('data2', 'PARANG', static=False)
        assert np.allclose(attribute1, attribute2, rtol=self.limit, atol=0.)
        assert np.allclose(attribute2, attribute3, rtol=self.limit, atol=0.)



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.processing.darkflat import DarkCalibrationModule, FlatCalibrationModule
from pynpoint.util.tests import create_config, remove_test_data


class TestDarkFlat:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        np.random.seed(1)

        images = np.random.normal(loc=0, scale=2e-4, size=(5, 11, 11))
        dark = np.random.normal(loc=0, scale=2e-4, size=(5, 11, 11))
        flat = np.random.normal(loc=0, scale=2e-4, size=(5, 11, 11))
        crop = np.random.normal(loc=0, scale=2e-4, size=(5, 7, 7))

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'w') as hdf_file:
            hdf_file.create_dataset('images', data=images)
            hdf_file.create_dataset('dark', data=dark)
            hdf_file.create_dataset('flat', data=flat)
            hdf_file.create_dataset('crop', data=crop)

        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir)

    def test_input_data(self) -> None:

        data = self.pipeline.get_data('dark')
        assert np.sum(data) == pytest.approx(-2.0262345764957305e-05, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

        data = self.pipeline.get_data('flat')
        assert np.sum(data) == pytest.approx(0.0076413379497053, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_dark_calibration(self) -> None:

        module = DarkCalibrationModule(name_in='dark',
                                       image_in_tag='images',
                                       dark_in_tag='dark',
                                       image_out_tag='dark_cal')

        self.pipeline.add_module(module)
        self.pipeline.run_module('dark')

        data = self.pipeline.get_data('dark_cal')
        assert np.sum(data) == pytest.approx(0.00717386583629883, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_flat_calibration(self) -> None:

        module = FlatCalibrationModule(name_in='flat',
                                       image_in_tag='dark_cal',
                                       flat_in_tag='flat',
                                       image_out_tag='flat_cal')

        self.pipeline.add_module(module)
        self.pipeline.run_module('flat')

        data = self.pipeline.get_data('flat_cal')
        assert np.sum(data) == pytest.approx(0.00717439711853594, rel=self.limit, abs=0.)
        assert data.shape == (5, 11, 11)

    def test_flat_crop(self) -> None:

        module = FlatCalibrationModule(name_in='flat_crop',
                                       image_in_tag='crop',
                                       flat_in_tag='flat',
                                       image_out_tag='flat_crop')

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('flat_crop')

        assert len(warning) == 1

        assert warning[0].message.args[0] == 'The calibration images were cropped around their ' \
                                             'center to match the shape of the science images.'

        data = self.pipeline.get_data('flat_crop')
        assert np.sum(data) == pytest.approx(-0.003242901413605404, rel=self.limit, abs=0.)
        assert data.shape == (5, 7, 7)

    def test_flat_too_small(self) -> None:

        module = FlatCalibrationModule(name_in='flat_small',
                                       image_in_tag='flat',
                                       flat_in_tag='crop',
                                       image_out_tag='flat_small')

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('flat_small')

        assert str(error.value) == 'Shape of the calibration images is smaller than the ' \
                                   'science images.'



"""
Module for reading FITS files obtained with VLT/VISIR for the NEAR experiment.
"""

import os
import math
import time
import shlex
import subprocess
import threading
import warnings

from typing import Optional, Union, Tuple

import numpy as np

from astropy.io import fits
from typeguard import typechecked

from pynpoint.core.processing import ReadingModule
from pynpoint.util.attributes import set_static_attr, set_nonstatic_attr, set_extra_attr
from pynpoint.util.module import progress, memory_frames
from pynpoint.util.image import crop_image


class NearReadingModule(ReadingModule):
    """
    Pipeline module for reading VLT/VISIR data of the NEAR experiment. The FITS files and required
    header information are read from the input directory and stored in two datasets, corresponding
    to chop A and chop B. The primary HDU of the FITS files should contain the main header
    information, while the subsequent HDUs contain each a single image (alternated for chop A and
    chop B) and some additional header information for that image. The last HDU is ignored as it
    contains the average of all images.
    """

    __author__ = 'Jasper Jonker, Tomas Stolker, Anna Boehle'

    @typechecked
    def __init__(self,
                 name_in: str,
                 input_dir: Optional[str] = None,
                 chopa_out_tag: str = 'chopa',
                 chopb_out_tag: str = 'chopb',
                 subtract: bool = False,
                 crop: Optional[Union[Tuple[int, int, float], Tuple[None, None, float]]] = None,
                 combine: Optional[str] = None):
        """
        Parameters
        ----------
        name_in : str
            Unique name of the instance.
        input_dir : str, None
            Input directory where the FITS files are located. The default input folder of the
            Pypeline is used if set to None.
        chopa_out_tag : str
            Database entry where the chop A images will be stored. Should be different from
            ``chop_b_out_tag``.
        chopb_out_tag : str
            Database entry where the chop B images will be stored. Should be different from
            ``chop_a_out_tag``.
        subtract : bool
            If True, the other chop position is subtracted before saving out the chop A and chop B
            images.
        crop: tuple(int, int, float), None
            The pixel position (x, y) around which the chop A and chop B images are cropped and
            the new image size (arcsec), together provided as (pos_x, pos_y, size). The same size
            will be used for both image dimensions. It is recommended to crop the images around
            the approximate coronagraph position. No cropping is applied if set to None.
        combine: str, None
            Method ('mean' or 'median') for combining (separately) the chop A and chop B frames
            from each cube into a single frame. All frames are stored if set to None.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in, input_dir=input_dir)

        self.m_chopa_out_port = self.add_output_port(chopa_out_tag)
        self.m_chopb_out_port = self.add_output_port(chopb_out_tag)

        self.m_subtract = subtract
        self.m_crop = crop
        self.m_combine = combine

    @staticmethod
    @typechecked
    def _uncompress_file(filename: str) -> None:
        """
        Internal function to uncompress a .Z file.

        Parameters
        ----------
        filename : str
            Compressed .Z file.

        Returns
        -------
        NoneType
            None
        """

        try:
            # try running a subprocess with the 'uncompress' command
            command = 'uncompress ' + filename
            subprocess.check_call(shlex.split(command))

        except(FileNotFoundError, OSError):
            # or else run a subprocess with the 'gunzip' command
            command = 'gunzip -d ' + filename
            subprocess.check_call(shlex.split(command))

    @typechecked
    def uncompress(self) -> None:
        """
        Method to check if the input directory contains compressed files ending with .fits.Z.
        If this is the case, the files will be uncompressed using multithreading. The number
        of threads can be set with the ``CPU`` parameter in the configuration file.

        Returns
        -------
        NoneType
            None
        """

        cpu = self._m_config_port.get_attribute('CPU')

        # list all files ending with .fits.Z in the input location
        files = []
        for item in os.listdir(self.m_input_location):
            if item.endswith('.fits.Z'):
                files.append(os.path.join(self.m_input_location, item))

        if files:
            # subdivide the file indices by number of CPU
            indices = memory_frames(cpu, len(files))

            start_time = time.time()
            for i, _ in enumerate(indices[:-1]):
                progress(i, len(indices[:-1]), 'Uncompressing NEAR data...', start_time)

                # select subset of compressed files
                subset = files[indices[i]:indices[i+1]]

                # create a list of threads to uncompress CPU number of files
                # each file is processed by a different thread
                threads = []
                for filename in subset:
                    thread = threading.Thread(target=self._uncompress_file, args=(filename, ))
                    threads.append(thread)

                # start the threads
                for item in threads:
                    item.start()

                # join the threads
                for item in threads:
                    item.join()

    @staticmethod
    @typechecked
    def check_header(header: fits.header.Header) -> None:
        """
        Method to check the header information and prompt a warning if a value is not as expected.

        Parameters
        ----------
        header : astropy.io.fits.header.Header
            Header information from the FITS file that is read.

        Returns
        -------
        NoneType
            None
        """

        if str(header['ESO DET CHOP ST']) == 'F':
            warnings.warn('Dataset was obtained without chopping.')

        skipped = int(header['ESO DET CHOP CYCSKIP'])
        if skipped != 0:
            warnings.warn(f'Chop cycles ({skipped}) have been skipped.')

        if str(header['ESO DET CHOP CYCSUM']) == 'T':
            warnings.warn('FITS file contains averaged images.')

    @typechecked
    def read_header(self,
                    filename: str) -> Tuple[fits.header.Header, Tuple[int, int, int]]:
        """
        Function that opens a FITS file and separates the chop A and chop B images. The primary HDU
        contains only a general header. The subsequent HDUs contain a single image with a small
        extra header. The last HDU is the average of all images, which will be ignored.

        Parameters
        ----------
        filename : str
            Absolute path and filename of the FITS file.

        Returns
        -------
        astropy.io.fits.header.Header
            Primary header, which is valid for all images.
        tuple(int, int, int)
            Shape of a stack of images for chop A or B.
        """

        # open the FITS file
        hdulist = fits.open(filename)

        # number of images = total number of HDUs - primary HDU - last HDU (average image)
        nimages = len(hdulist) - 2

        # check if the file contains an even number of images, as expected with two chop positions
        if nimages % 2 != 0:
            warnings.warn(f'FITS file contains odd number of images: {filename}')

            # decreasing nimages to an even number such that nimages // 2 gives the correct size
            nimages -= 1

        # primary header
        header = hdulist[0].header

        # number of chop cycles
        ncycles = header['ESO DET CHOP NCYCLES']

        # number of chop cycles should be equal to half the number of available images
        if ncycles != nimages // 2:
            warnings.warn(f'The number of chop cycles ({ncycles}) is not equal to half the '
                          f'number of available HDU images ({nimages // 2}).')

        # header of the first image
        header_image = hdulist[1].header

        # create a list of key = value from the primary header
        fits_header = []
        for key in header:
            if key:
                fits_header.append(str(key)+' = '+str(header[key]))

        # write the primary header information to the fits_header group
        header_out_port = self.add_output_port('fits_header/' + filename)
        header_out_port.set_all(np.array(fits_header))

        # shape of the image stacks for chop A/B (hence nimages/2)
        im_shape = (nimages // 2, header_image['NAXIS2'], header_image['NAXIS1'])

        # set the NAXIS image shape in the primary header
        # required by util.attributes.set_nonstatic_attr
        header.set('NAXIS', 3)
        header.set('NAXIS1', im_shape[2])
        header.set('NAXIS2', im_shape[1])
        header.set('NAXIS3', im_shape[0])

        # check primary header
        self.check_header(header)

        hdulist.close()

        return header, im_shape

    @staticmethod
    @typechecked
    def read_images(filename: str,
                    im_shape: Tuple[int, int, int]) -> Tuple[np.ndarray, np.ndarray]:
        """
        Function that opens a FITS file and separates the chop A and chop B images. The primary HDU
        contains only a general header. The subsequent HDUs contain a single image with a small
        extra header. The last HDU is the average of all images, which will be ignored.

        Parameters
        ----------
        filename : str
            Absolute path and filename of the FITS file.
        im_shape : tuple(int, int, int)
            Shape of a stack of images for chop A or B.

        Returns
        -------
        numpy.array
            Array containing the images of chop A.
        numpy.array
            Array containing the images of chop B.
        """

        # open the FITS file
        hdulist = fits.open(filename)

        # initialize the image arrays for chop A and B
        chopa = np.zeros(im_shape, dtype=np.float32)
        chopb = np.zeros(im_shape, dtype=np.float32)

        count_chopa, count_chopb = 0, 0
        prev_cycle = None

        for i in range(2*im_shape[0]):
            # get the chop position (HCYCLE1 = chop A, HCYCLE2 = chop B)
            # primary HDU is skipped with +1
            if 'ESO DET FRAM TYPE' in hdulist[i+1].header:
                cycle = hdulist[i+1].header['ESO DET FRAM TYPE']

            else:
                hdulist.close()
                raise ValueError(f'Frame type not found in the FITS header. Image number: {i}.')

            # write the HDU image to the chop A or B array
            # count the number of chop A and B images
            if cycle == 'HCYCLE1' and cycle != prev_cycle:
                chopa[count_chopa, ] = hdulist[i+1].data.byteswap().newbyteorder()

                count_chopa += 1
                prev_cycle = cycle

            elif cycle == 'HCYCLE2' and cycle != prev_cycle:
                chopb[count_chopb, ] = hdulist[i+1].data.byteswap().newbyteorder()

                count_chopb += 1
                prev_cycle = cycle

            elif cycle == prev_cycle:
                warnings.warn(f'Previous and current chop position ({cycle}) are the same. '
                              'Skipping the current image.')

            else:
                hdulist.close()
                raise ValueError(f'Frame type ({cycle}) not a valid value. Expecting HCYCLE1 or '
                                 'HCYCLE2 as value for ESO DET FRAM TYPE.')

        # check if the number of chop A and B images is equal, this error should never occur
        if count_chopa != count_chopb:
            warnings.warn('The number of images is not equal for chop A and chop B.')

        hdulist.close()

        return chopa, chopb

    @typechecked
    def run(self) -> None:
        """
        Run the module. The FITS files are collected from the input directory and uncompressed if
        needed. The images are then sorted by the two chop positions (chop A and chop B). The
        required FITS header keywords (which should be set in the configuration file) are also
        imported and stored as attributes to the two output datasets in the HDF5 database.

        Returns
        -------
        NoneType
            None
        """

        # clear the output ports
        self.m_chopa_out_port.del_all_data()
        self.m_chopa_out_port.del_all_attributes()
        self.m_chopb_out_port.del_all_data()
        self.m_chopb_out_port.del_all_attributes()

        # uncompress the FITS files if needed
        self.uncompress()

        # find and sort the FITS files
        files = []

        for filename in os.listdir(self.m_input_location):
            if filename.endswith('.fits'):
                files.append(os.path.join(self.m_input_location, filename))

        files.sort()

        # check if there are FITS files present in the input location
        assert files, f'No FITS files found in {self.m_input_location}.'

        # if cropping chop A, get pixscale and convert crop_size to pixels and swap x/y
        if self.m_crop is not None:
            pixscale = self._m_config_port.get_attribute('PIXSCALE')
            self.m_crop = (self.m_crop[1], self.m_crop[0], int(math.ceil(self.m_crop[2]/pixscale)))

        start_time = time.time()
        for i, filename in enumerate(files):
            progress(i, len(files), 'Preprocessing NEAR data...', start_time)

            # get the primary header data and the image shape
            header, im_shape = self.read_header(filename)

            # get the images of chop A and chop B
            chopa, chopb = self.read_images(filename, im_shape)

            if self.m_subtract:
                chopa = chopa - chopb
                chopb = -1. * np.copy(chopa)

            if self.m_crop is not None:
                chopa = crop_image(chopa,
                                   center=self.m_crop[0:2],
                                   size=self.m_crop[2],
                                   copy=False)

                chopb = crop_image(chopb,
                                   center=self.m_crop[0:2],
                                   size=self.m_crop[2],
                                   copy=False)

            if self.m_combine is not None:

                if self.m_combine == 'mean':
                    chopa = np.mean(chopa, axis=0)
                    chopb = np.mean(chopb, axis=0)

                elif self.m_combine == 'median':
                    chopa = np.median(chopa, axis=0)
                    chopb = np.median(chopb, axis=0)

                header[self._m_config_port.get_attribute('NFRAMES')] = 1

            # append the images of chop A and B
            self.m_chopa_out_port.append(chopa, data_dim=3)
            self.m_chopb_out_port.append(chopb, data_dim=3)

            # starting value for the INDEX attribute
            first_index = 0

            for port in (self.m_chopa_out_port, self.m_chopb_out_port):

                # set the static attributes
                set_static_attr(fits_file=filename,
                                header=header,
                                config_port=self._m_config_port,
                                image_out_port=port,
                                check=True)

                # set the non-static attributes
                set_nonstatic_attr(header=header,
                                   config_port=self._m_config_port,
                                   image_out_port=port,
                                   check=True)

                # set the remaining attributes
                set_extra_attr(fits_file=filename,
                               nimages=im_shape[0]//2,
                               config_port=self._m_config_port,
                               image_out_port=port,
                               first_index=first_index)

                # increase the first value of the INDEX attribute
                first_index += im_shape[0]//2

                # flush the output port
                port.flush()

        # add history information
        self.m_chopa_out_port.add_history('NearReadingModule', 'Chop A')
        self.m_chopb_out_port.add_history('NearReadingModule', 'Chop B')

        # close all connections to the database
        self.m_chopa_out_port.close_port()



import os

import h5py
import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.fitsreading import FitsReadingModule
from pynpoint.processing.fluxposition import FakePlanetModule, AperturePhotometryModule, \
                                             FalsePositiveModule, SimplexMinimizationModule, \
                                             MCMCsamplingModule, SystematicErrorModule
from pynpoint.processing.stacksubset import DerotateAndStackModule
from pynpoint.processing.psfsubtraction import PcaPsfSubtractionModule
from pynpoint.util.tests import create_config, create_star_data, create_fake_data, remove_test_data


class TestFluxPosition:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_fake_data(self.test_dir+'adi')
        create_star_data(self.test_dir+'psf', npix=21, pos_star=10.)
        create_star_data(self.test_dir+'ref', npix=21, pos_star=10.)
        create_config(self.test_dir+'PynPoint_config.ini')

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, folders=['adi', 'psf', 'ref'])

    def test_read_data(self) -> None:

        module = FitsReadingModule(name_in='read1',
                                   image_tag='adi',
                                   input_dir=self.test_dir+'adi')

        self.pipeline.add_module(module)
        self.pipeline.run_module('read1')

        data = self.pipeline.get_data('adi')
        assert np.sum(data) == pytest.approx(11.012854046962481, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

        self.pipeline.set_attribute('adi', 'PARANG', np.linspace(0., 180., 10), static=False)

        module = FitsReadingModule(name_in='read2',
                                   image_tag='psf',
                                   input_dir=self.test_dir+'psf')

        self.pipeline.add_module(module)
        self.pipeline.run_module('read2')

        data = self.pipeline.get_data('psf')
        assert np.sum(data) == pytest.approx(108.43655133957289, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

        module = FitsReadingModule(name_in='read3',
                                   image_tag='ref',
                                   input_dir=self.test_dir+'psf')

        self.pipeline.add_module(module)
        self.pipeline.run_module('read3')

        data = self.pipeline.get_data('ref')
        assert np.sum(data) == pytest.approx(108.43655133957289, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

    def test_aperture_photometry(self) -> None:

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 1

        module = AperturePhotometryModule(name_in='photometry1',
                                          image_in_tag='psf',
                                          phot_out_tag='photometry1',
                                          radius=0.1,
                                          position=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('photometry1')

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 4

        module = AperturePhotometryModule(name_in='photometry2',
                                          image_in_tag='psf',
                                          phot_out_tag='photometry2',
                                          radius=0.1,
                                          position=None)

        self.pipeline.add_module(module)
        self.pipeline.run_module('photometry2')

        data = self.pipeline.get_data('photometry1')
        assert np.sum(data) == pytest.approx(100.80648929590365, rel=self.limit, abs=0.)
        assert data.shape == (10, 1)

        data_multi = self.pipeline.get_data('photometry2')
        assert data.shape == data_multi.shape
        assert data == pytest.approx(data_multi, rel=self.limit, abs=0.)

    def test_aperture_photometry_position(self) -> None:

        module = AperturePhotometryModule(name_in='photometry3',
                                          image_in_tag='psf',
                                          phot_out_tag='photometry3',
                                          radius=0.1,
                                          position=(10., 10.))

        self.pipeline.add_module(module)
        self.pipeline.run_module('photometry3')

        data = self.pipeline.get_data('photometry3')
        assert np.sum(data) == pytest.approx(100.80648929590365, rel=self.limit, abs=0.)
        assert data.shape == (10, 1)

    def test_fake_planet(self) -> None:

        module = FakePlanetModule(position=(0.2, 180.),
                                  magnitude=2.5,
                                  psf_scaling=1.,
                                  interpolation='spline',
                                  name_in='fake',
                                  image_in_tag='adi',
                                  psf_in_tag='psf',
                                  image_out_tag='fake')

        self.pipeline.add_module(module)
        self.pipeline.run_module('fake')

        data = self.pipeline.get_data('fake')
        assert np.sum(data) == pytest.approx(21.273233520675586, rel=self.limit, abs=0.)
        assert data.shape == (10, 21, 21)

    def test_psf_subtraction(self) -> None:

        module = PcaPsfSubtractionModule(pca_numbers=[1, ],
                                         name_in='pca',
                                         images_in_tag='fake',
                                         reference_in_tag='fake',
                                         res_mean_tag='res_mean',
                                         extra_rot=0.)

        self.pipeline.add_module(module)
        self.pipeline.run_module('pca')

        data = self.pipeline.get_data('res_mean')
        assert np.sum(data) == pytest.approx(0.013659056187572433, rel=self.limit, abs=0.)
        assert data.shape == (1, 21, 21)

    def test_false_positive(self) -> None:

        module = FalsePositiveModule(position=(10., 2.),
                                     aperture=0.06,
                                     ignore=True,
                                     name_in='false1',
                                     image_in_tag='res_mean',
                                     snr_out_tag='snr_fpf1',
                                     optimize=False)

        self.pipeline.add_module(module)
        self.pipeline.run_module('false1')

        data = self.pipeline.get_data('snr_fpf1')
        assert data[0, 1] == pytest.approx(2., rel=self.limit, abs=0.)
        assert data[0, 2] == pytest.approx(0.216, rel=self.limit, abs=0.)
        assert data[0, 3] == pytest.approx(180., rel=self.limit, abs=0.)
        assert data[0, 4] == pytest.approx(26.661611583224417, rel=self.limit, abs=0.)
        assert data[0, 5] == pytest.approx(1.3375156927387672e-08, rel=self.limit, abs=0.)
        assert data.shape == (1, 6)

    def test_false_positive_optimize(self) -> None:

        module = FalsePositiveModule(position=(10., 2.),
                                     aperture=0.06,
                                     ignore=True,
                                     name_in='false2',
                                     image_in_tag='res_mean',
                                     snr_out_tag='snr_fpf2',
                                     optimize=True,
                                     offset=0.1,
                                     tolerance=0.01)

        self.pipeline.add_module(module)
        self.pipeline.run_module('false2')

        data = self.pipeline.get_data('snr_fpf2')
        assert data[0, 1] == pytest.approx(2.0747802734374985, rel=self.limit, abs=0.)
        assert data[0, 2] == pytest.approx(0.2139883890923524, rel=self.limit, abs=0.)
        assert data[0, 3] == pytest.approx(179.52168877335356, rel=self.limit, abs=0.)
        assert data[0, 4] == pytest.approx(27.457328210661814, rel=self.limit, abs=0.)
        assert data[0, 5] == pytest.approx(1.0905578015907869e-08, rel=self.limit, abs=0.)
        assert data.shape == (1, 6)

    def test_simplex_minimization_hessian(self) -> None:

        module = SimplexMinimizationModule(name_in='simplex1',
                                           image_in_tag='fake',
                                           psf_in_tag='psf',
                                           res_out_tag='simplex_res',
                                           flux_position_tag='flux_position',
                                           position=(10., 3.),
                                           magnitude=2.5,
                                           psf_scaling=-1.,
                                           merit='hessian',
                                           aperture=0.06,
                                           sigma=0.,
                                           tolerance=0.1,
                                           pca_number=1,
                                           cent_size=0.06,
                                           edge_size=None,
                                           extra_rot=0.,
                                           reference_in_tag=None,
                                           residuals='median',
                                           offset=1.)

        self.pipeline.add_module(module)
        self.pipeline.run_module('simplex1')

        data = self.pipeline.get_data('simplex_res')

        assert np.sum(data) == pytest.approx(0.2781582369128238, rel=self.limit, abs=0.)
        assert data.shape == (35, 21, 21)

        data = self.pipeline.get_data('flux_position')
        assert data[24, 0] == pytest.approx(9.931627229080938, rel=self.limit, abs=0.)
        assert data[24, 1] == pytest.approx(2.6575231481481456, rel=self.limit, abs=0.)
        assert data[24, 2] == pytest.approx(0.1982554700445013, rel=self.limit, abs=0.)
        assert data[24, 3] == pytest.approx(179.46648003649148, rel=self.limit, abs=0.)
        assert data[24, 4] == pytest.approx(2.5256451474622708, rel=self.limit, abs=0.)
        assert data.shape == (35, 6)

    def test_simplex_minimization_reference(self) -> None:

        module = SimplexMinimizationModule(name_in='simplex2',
                                           image_in_tag='fake',
                                           psf_in_tag='psf',
                                           res_out_tag='simplex_res_ref',
                                           flux_position_tag='flux_position_ref',
                                           position=(10., 3.),
                                           magnitude=2.5,
                                           psf_scaling=-1.,
                                           merit='poisson',
                                           aperture=0.06,
                                           sigma=0.,
                                           tolerance=0.1,
                                           pca_number=1,
                                           cent_size=0.06,
                                           edge_size=None,
                                           extra_rot=0.,
                                           reference_in_tag='ref',
                                           residuals='mean')

        self.pipeline.add_module(module)
        self.pipeline.run_module('simplex2')

        data = self.pipeline.get_data('simplex_res_ref')
        assert np.sum(data) == pytest.approx(9.734993454838076, rel=self.limit, abs=0.)
        assert data.shape == (28, 21, 21)

        data = self.pipeline.get_data('flux_position_ref')
        assert data[27, 0] == pytest.approx(10.049019964116436, rel=self.limit, abs=0.)
        assert data[27, 1] == pytest.approx(2.6444836362361936, rel=self.limit, abs=0.)
        assert data[27, 2] == pytest.approx(0.19860335205689572, rel=self.limit, abs=0.)
        assert data[27, 3] == pytest.approx(180.38183525629643, rel=self.limit, abs=0.)
        assert data[27, 4] == pytest.approx(2.5496922175196, rel=self.limit, abs=0.)
        assert data.shape == (28, 6)

    def test_mcmc_sampling(self) -> None:

        with h5py.File(self.test_dir+'PynPoint_database.hdf5', 'a') as hdf_file:
            hdf_file['config'].attrs['CPU'] = 4

        module = DerotateAndStackModule(name_in='stack',
                                        image_in_tag='psf',
                                        image_out_tag='psf_stack',
                                        derotate=False,
                                        stack='mean')

        self.pipeline.add_module(module)
        self.pipeline.run_module('stack')

        data = self.pipeline.get_data('psf_stack')
        assert np.sum(data) == pytest.approx(10.843655133957288, rel=self.limit, abs=0.)
        assert data.shape == (1, 21, 21)

        module = MCMCsamplingModule(name_in='mcmc1',
                                    image_in_tag='adi',
                                    psf_in_tag='psf_stack',
                                    chain_out_tag='mcmc',
                                    param=(0.15, 0., 1.),
                                    bounds=((0.1, 0.2), (-2., 2.), (-1., 2.)),
                                    nwalkers=6,
                                    nsteps=5,
                                    psf_scaling=-1.,
                                    pca_number=1,
                                    aperture=(10, 16, 0.06),
                                    mask=None,
                                    extra_rot=0.,
                                    merit='gaussian',
                                    residuals='median',
                                    resume=False,
                                    sigma=(1e-3, 1e-1, 1e-2))

        self.pipeline.add_module(module)
        self.pipeline.run_module('mcmc1')

        data = self.pipeline.get_data('mcmc')
        assert data.shape == (5, 6, 3)

        data = self.pipeline.get_data('mcmc_backend')
        assert data.shape == (3, )

        module = MCMCsamplingModule(name_in='mcmc2',
                                    image_in_tag='adi',
                                    psf_in_tag='psf_stack',
                                    chain_out_tag='mcmc',
                                    param=(0.15, 0., 1.),
                                    bounds=((0.1, 0.2), (-2., 2.), (-1., 2.)),
                                    nwalkers=6,
                                    nsteps=5,
                                    psf_scaling=-1.,
                                    pca_number=1,
                                    aperture=(10, 16, 0.06),
                                    mask=None,
                                    extra_rot=0.,
                                    merit='gaussian',
                                    residuals='median',
                                    resume=True,
                                    sigma=(1e-3, 1e-1, 1e-2))

        self.pipeline.add_module(module)
        self.pipeline.run_module('mcmc2')

        data = self.pipeline.get_data('mcmc')
        assert data.shape == (10, 6, 3)

        data = self.pipeline.get_data('mcmc_backend')
        assert data.shape == (3, )

    def test_systematic_error(self) -> None:

        module = SystematicErrorModule(name_in='error',
                                       image_in_tag='adi',
                                       psf_in_tag='psf',
                                       offset_out_tag='offset',
                                       position=(0.162, 0.),
                                       magnitude=5.,
                                       angles=(0., 180., 2),
                                       psf_scaling=1.,
                                       merit='gaussian',
                                       aperture=0.06,
                                       tolerance=0.1,
                                       pca_number=1,
                                       mask=(None, None),
                                       extra_rot=0.,
                                       residuals='median',
                                       offset=1.)

        self.pipeline.add_module(module)
        self.pipeline.run_module('error')

        data = self.pipeline.get_data('offset')
        assert data[0, 0] == pytest.approx(-0.001114020093541973, rel=self.limit, abs=0.)
        assert data[0, 1] == pytest.approx(-0.012163271644183737, rel=self.limit, abs=0.)
        assert data[0, 2] == pytest.approx(-0.017943854263249293, rel=self.limit, abs=0.)
        assert data[0, 3] == pytest.approx(0.001282493868968615, rel=self.limit, abs=0.)
        assert data[0, 4] == pytest.approx(-0.04125986733475884, rel=self.limit, abs=0.)
        assert data.shape == (2, 5)



"""
Module for reading HDF5 files that were created with the
:class:`~pynpoint.readwrite.hdf5writing.Hdf5WritingModule`.
"""

import os
import time
import warnings

from typing import Optional

import h5py
import numpy as np

from typeguard import typechecked

from pynpoint.core.processing import ReadingModule
from pynpoint.util.module import progress


class Hdf5ReadingModule(ReadingModule):
    """
    Reads an HDF5 file from the given *input_dir* or the default directory of the Pypeline. A tag
    dictionary has to be set in order to choose the datasets which will be imported into the
    database. Also the static and non-static attributes are read from the HDF5 file and stored
    in the database with the corresponding data set. This module should only be used for reading
    HDF5 files that are created with the Hdf5WritingModule. Reading different type of HDF5 files
    may lead to inconsistencies in the central database.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 input_filename: Optional[str] = None,
                 input_dir: Optional[str] = None,
                 tag_dictionary: Optional[dict] = None):
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        input_filename : str, None
            The file name of the HDF5 input file. All files inside the input location will be
            imported if no filename is provided.
        input_dir : str, None
            The directory of the input HDF5 file. If no location is given, the default input
            location of the Pypeline is used.
        tag_dictionary : dict, None
            Dictionary of all data sets that will be imported. The dictionary format is
            {*tag_name_in_input_file*:*tag_name_in_database*, }. All data sets in the input HDF5
            file that match one of the *tag_name_in_input_file* will be imported. The tag name
            inside the internal Pypeline database will be changed to *tag_name_in_database*.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in, input_dir=input_dir)

        if tag_dictionary is None:
            tag_dictionary = {}

        for out_tag in tag_dictionary.values():
            self.add_output_port(out_tag)

        self.m_filename = input_filename
        self._m_tag_dictionary = tag_dictionary

    @typechecked
    def read_single_hdf5(self,
                         file_in: str) -> None:
        """
        Function which reads a single HDF5 file.

        Parameters
        ----------
        file_in : str
            Path and name of the HDF5 file.

        Returns
        -------
        NoneType
            None
        """

        hdf5_file = h5py.File(file_in, mode='r')

        for tag_in in self._m_tag_dictionary:
            tag_in = str(tag_in)  # unicode keys cause errors
            tag_out = self._m_tag_dictionary[tag_in]

            if tag_in not in hdf5_file:
                warnings.warn(f'The dataset with tag name \'{tag_in}\' is not found in the HDF5 '
                              f'file.')

                continue

            # add data
            port_out = self._m_output_ports[tag_out]
            port_out.set_all(np.asarray(hdf5_file[tag_in][...]))

            # add static attributes
            for attr_name, attr_value in hdf5_file[tag_in].attrs.items():
                port_out.add_attribute(name=attr_name, value=attr_value)

            # add non-static attributes
            if 'header_' + tag_in in hdf5_file:
                for attr_name in hdf5_file['header_' + tag_in]:
                    attr_val = hdf5_file['header_' + tag_in + '/' + attr_name][...]
                    port_out.add_attribute(name=attr_name, value=attr_val, static=False)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Looks for all HDF5 files in the input directory and reads the
        datasets that are provided in the tag dictionary.

        Returns
        -------
        NoneType
            None
        """

        # create list of files to be read
        files = []

        tmp_dir = os.path.join(self.m_input_location, '')

        # check if a single input file is given
        if self.m_filename is not None:
            # create file path + filename
            assert(os.path.isfile((tmp_dir + str(self.m_filename)))), \
                   f'Error: Input file does not exist. Input requested: {self.m_filename}'

            files.append((tmp_dir + str(self.m_filename)))

        else:
            # look for all HDF5 files in the directory
            for tmp_file in os.listdir(self.m_input_location):
                if tmp_file.endswith('.hdf5') or tmp_file.endswith('.h5'):
                    files.append(tmp_dir + str(tmp_file))

        start_time = time.time()
        for i, tmp_file in enumerate(files):
            progress(i, len(files), 'Reading HDF5 file...', start_time)
            self.read_single_hdf5(tmp_file)



import os

import pytest
import numpy as np

from pynpoint.core.pypeline import Pypeline
from pynpoint.readwrite.attr_reading import ParangReadingModule, AttributeReadingModule, \
                                            WavelengthReadingModule
from pynpoint.util.tests import create_config, create_random, remove_test_data


class TestAttributeReading:

    def setup_class(self) -> None:

        self.limit = 1e-10
        self.test_dir = os.path.dirname(__file__) + '/'

        create_random(self.test_dir, nimages=10)
        create_config(self.test_dir+'PynPoint_config.ini')

        np.savetxt(self.test_dir+'parang.dat', np.arange(10., 20., 1.))
        np.savetxt(self.test_dir+'new.dat', np.arange(20., 30., 1.))
        np.savetxt(self.test_dir+'attribute.dat', np.arange(0, 10, 1), fmt='%i')
        np.savetxt(self.test_dir+'wavelength.dat', np.arange(0., 10., 1.))

        data2d = np.random.normal(loc=0, scale=2e-4, size=(10, 10))
        np.savetxt(self.test_dir+'data_2d.dat', data2d)

        self.pipeline = Pypeline(self.test_dir, self.test_dir, self.test_dir)

    def teardown_class(self) -> None:

        remove_test_data(self.test_dir, files=['parang.dat', 'new.dat',
                                               'attribute.dat', 'data_2d.dat',
                                               'wavelength.dat'])

    def test_input_data(self) -> None:

        data = self.pipeline.get_data('images')
        assert np.sum(data) == pytest.approx(0.007133341144768919, rel=self.limit, abs=0.)
        assert data.shape == (10, 11, 11)

    def test_parang_reading(self) -> None:

        module = ParangReadingModule(name_in='parang1',
                                     data_tag='images',
                                     file_name='parang.dat',
                                     input_dir=None,
                                     overwrite=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('parang1')

        data = self.pipeline.get_data('header_images/PARANG')
        assert data.dtype == 'float64'
        assert data == pytest.approx(np.arange(10., 20., 1.), rel=self.limit, abs=0.)
        assert data.shape == (10, )

    def test_parang_reading_same(self) -> None:

        module = ParangReadingModule(name_in='parang2',
                                     data_tag='images',
                                     file_name='parang.dat',
                                     input_dir=None,
                                     overwrite=True)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('parang2')

        assert len(warning) == 1
        assert warning[0].message.args[0] == 'The PARANG attribute is already present and ' \
                                             'contains the same values as are present in ' \
                                             'parang.dat.'

    def test_parang_reading_present(self) -> None:

        module = ParangReadingModule(name_in='parang3',
                                     data_tag='images',
                                     file_name='new.dat',
                                     input_dir=None,
                                     overwrite=False)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('parang3')

        assert len(warning) == 1
        assert warning[0].message.args[0] == 'The PARANG attribute is already present. Set the ' \
                                             '\'overwrite\' parameter to True in order to ' \
                                             'overwrite the values with new.dat.'

    def test_parang_reading_overwrite(self) -> None:

        module = ParangReadingModule(file_name='new.dat',
                                     name_in='parang4',
                                     input_dir=None,
                                     data_tag='images',
                                     overwrite=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('parang4')

    def test_parang_reading_2d(self) -> None:

        module = ParangReadingModule(name_in='parang6',
                                     data_tag='images',
                                     file_name='data_2d.dat',
                                     input_dir=None,
                                     overwrite=False)

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('parang6')

        assert str(error.value) == 'The input file data_2d.dat should contain a 1D data set with ' \
                                   'the parallactic angles.'

    def test_attribute_reading(self) -> None:

        module = AttributeReadingModule(file_name='attribute.dat',
                                        attribute='EXP_NO',
                                        name_in='attribute1',
                                        input_dir=None,
                                        data_tag='images',
                                        overwrite=False)

        self.pipeline.add_module(module)
        self.pipeline.run_module('attribute1')

        data = self.pipeline.get_data('header_images/EXP_NO')
        assert data.dtype == 'int64'
        assert data == pytest.approx(np.arange(10), rel=self.limit, abs=0.)
        assert data.shape == (10, )

    def test_attribute_reading_present(self) -> None:

        module = AttributeReadingModule(file_name='parang.dat',
                                        attribute='PARANG',
                                        name_in='attribute3',
                                        input_dir=None,
                                        data_tag='images',
                                        overwrite=False)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('attribute3')

        assert warning[0].message.args[0] == 'The attribute \'PARANG\' is already present. Set ' \
                                             'the \'overwrite\' parameter to True in order to ' \
                                             'overwrite the values with parang.dat.'

    def test_attribute_reading_invalid(self) -> None:

        module = AttributeReadingModule(file_name='attribute.dat',
                                        attribute='test',
                                        name_in='attribute4',
                                        input_dir=None,
                                        data_tag='images',
                                        overwrite=False)

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('attribute4')

        assert str(error.value) == '\'test\' is not a valid attribute.'

    def test_attribute_reading_2d(self) -> None:

        module = AttributeReadingModule(file_name='data_2d.dat',
                                        attribute='DITHER_X',
                                        name_in='attribute5',
                                        input_dir=None,
                                        data_tag='images',
                                        overwrite=False)

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('attribute5')

        assert str(error.value) == 'The input file data_2d.dat should contain a 1D list with ' \
                                   'attributes.'

    def test_attribute_reading_same(self) -> None:

        module = AttributeReadingModule(file_name='attribute.dat',
                                        attribute='EXP_NO',
                                        name_in='attribute6',
                                        input_dir=None,
                                        data_tag='images',
                                        overwrite=True)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('attribute6')

        assert len(warning) == 1
        assert warning[0].message.args[0] == 'The \'EXP_NO\' attribute is already present and ' \
                                             'contains the same values as are present in ' \
                                             'attribute.dat.'

    def test_attribute_reading_overwrite(self) -> None:

        module = AttributeReadingModule(file_name='parang.dat',
                                        attribute='PARANG',
                                        name_in='attribute7',
                                        input_dir=None,
                                        data_tag='images',
                                        overwrite=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('attribute7')

        data = self.pipeline.get_attribute('images', 'PARANG', static=False)
        assert data == pytest.approx(np.arange(10., 20., 1.), rel=self.limit, abs=0.)

    def test_wavelength_reading(self) -> None:

        module = WavelengthReadingModule(file_name='wavelength.dat',
                                         name_in='wavelength1',
                                         input_dir=None,
                                         data_tag='images',
                                         overwrite=False)

        self.pipeline.add_module(module)
        self.pipeline.run_module('wavelength1')

        data = self.pipeline.get_data('header_images/WAVELENGTH')
        assert data.dtype == 'float64'
        assert data == pytest.approx(np.arange(10.), rel=self.limit, abs=0.)
        assert data.shape == (10, )

    def test_wavelength_reading_same(self) -> None:

        module = WavelengthReadingModule(file_name='wavelength.dat',
                                         name_in='wavelength2',
                                         input_dir=None,
                                         data_tag='images',
                                         overwrite=True)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('wavelength2')

        assert len(warning) == 1
        assert warning[0].message.args[0] == 'The WAVELENGTH attribute is already present and ' \
                                             'contains the same values as are present in ' \
                                             'wavelength.dat.'

    def test_wavelength_reading_present(self) -> None:

        module = WavelengthReadingModule(file_name='new.dat',
                                         name_in='wavelength3',
                                         input_dir=None,
                                         data_tag='images',
                                         overwrite=False)

        self.pipeline.add_module(module)

        with pytest.warns(UserWarning) as warning:
            self.pipeline.run_module('wavelength3')

        assert len(warning) == 1
        assert warning[0].message.args[0] == 'The WAVELENGTH attribute is already present. Set ' \
                                             'the \'overwrite\' parameter to True in order to ' \
                                             'overwrite the values with new.dat.'

    def test_wavelength_reading_overwrite(self) -> None:

        module = WavelengthReadingModule(file_name='new.dat',
                                         name_in='wavelength4',
                                         input_dir=None,
                                         data_tag='images',
                                         overwrite=True)

        self.pipeline.add_module(module)
        self.pipeline.run_module('wavelength4')

    def test_wavelength_reading_2d(self) -> None:

        module = WavelengthReadingModule(file_name='data_2d.dat',
                                         name_in='wavelength6',
                                         input_dir=None,
                                         data_tag='images',
                                         overwrite=False)

        self.pipeline.add_module(module)

        with pytest.raises(ValueError) as error:
            self.pipeline.run_module('wavelength6')

        assert str(error.value) == 'The input file data_2d.dat should contain a 1D data set with ' \
                                   'the wavelengths.'



"""
Pipeline modules for dark frame and flat field calibrations.
"""

import time
import warnings

from typing import Tuple

import numpy as np

from typeguard import typechecked

from pynpoint.core.processing import ProcessingModule
from pynpoint.util.module import progress, memory_frames


@typechecked
def _master_frame(data: np.ndarray,
                  im_shape: Tuple[int, int, int]) -> np.ndarray:
    """
    Internal function which creates a master dark/flat by calculating the mean (3D data) and
    cropping the frames to the shape of the science images if needed.

    Parameters
    ----------
    data : numpy.ndarray
        Input array (2D) with mean of the dark or flat frames.
    im_shape : tuple(int, int, int)
        Shape of the science images (3D).

    Returns
    -------
    numpy.ndarray
        Master dark/flat frame.
    """

    shape_in = (im_shape[1], im_shape[2])

    if data.shape[0] < shape_in[0] or data.shape[1] < shape_in[1]:
        raise ValueError('Shape of the calibration images is smaller than the science images.')

    if data.shape != shape_in:
        cal_shape = data.shape

        x_off = (cal_shape[0] - shape_in[0]) // 2
        y_off = (cal_shape[1] - shape_in[1]) // 2

        data = data[x_off:x_off+shape_in[0], y_off:y_off+shape_in[1]]

        warnings.warn('The calibration images were cropped around their center to match the shape '
                      'of the science images.')

    return data


class DarkCalibrationModule(ProcessingModule):
    """
    Pipeline module to subtract a master dark from the science data.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 dark_in_tag: str,
                 image_out_tag: str) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the database entry with the science images that are read as input.
        dark_in_tag : str
            Tag of the database with the dark frames that are read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_dark_in_port = self.add_input_port(dark_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Creates a master dark with the same shape as the science
        data and subtracts the dark frame from the science data.

        Returns
        -------
        NoneType
            None
        """

        memory = self._m_config_port.get_attribute('MEMORY')
        nimages = self.m_image_in_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        dark = self.m_dark_in_port.get_all()

        master = _master_frame(data=np.mean(dark, axis=0),
                               im_shape=self.m_image_in_port.get_shape())

        start_time = time.time()

        for i in range(len(frames[:-1])):
            progress(i, len(frames[:-1]), 'Subtracting the dark current...', start_time)

            images = self.m_image_in_port[frames[i]:frames[i+1], ]

            self.m_image_out_port.append(images - master, data_dim=3)

        history = f'dark_in_tag = {self.m_dark_in_port.tag}'
        self.m_image_out_port.add_history('DarkCalibrationModule', history)
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.close_port()


class FlatCalibrationModule(ProcessingModule):
    """
    Pipeline module to apply a flat field correction to the science data.
    """

    __author__ = 'Markus Bonse, Tomas Stolker'

    @typechecked
    def __init__(self,
                 name_in: str,
                 image_in_tag: str,
                 flat_in_tag: str,
                 image_out_tag: str) -> None:
        """
        Parameters
        ----------
        name_in : str
            Unique name of the module instance.
        image_in_tag : str
            Tag of the science database that is read as input.
        flat_in_tag : str
            Tag of the flat field database that is read as input.
        image_out_tag : str
            Tag of the database entry that is written as output.

        Returns
        -------
        NoneType
            None
        """

        super().__init__(name_in)

        self.m_image_in_port = self.add_input_port(image_in_tag)
        self.m_flat_in_port = self.add_input_port(flat_in_tag)
        self.m_image_out_port = self.add_output_port(image_out_tag)

    @typechecked
    def run(self) -> None:
        """
        Run method of the module. Creates a master flat with the same shape as the science
        image and divides the science images by the flat field.

        Returns
        -------
        NoneType
            None
        """

        memory = self._m_config_port.get_attribute('MEMORY')
        nimages = self.m_image_in_port.get_shape()[0]
        frames = memory_frames(memory, nimages)

        flat = self.m_flat_in_port.get_all()

        master = _master_frame(data=np.mean(flat, axis=0),
                               im_shape=self.m_image_in_port.get_shape())

        # shift all values to greater or equal to +1.0
        flat_min = np.amin(master)
        master -= flat_min - 1.

        # normalization, median value is 1 afterwards
        master /= np.median(master)

        start_time = time.time()

        for i in range(len(frames[:-1])):
            progress(i, len(frames[:-1]), 'Flat fielding the images...', start_time)

            images = self.m_image_in_port[frames[i]:frames[i+1], ]

            self.m_image_out_port.append(images/master, data_dim=3)

        history = f'flat_in_tag = {self.m_flat_in_port.tag}'
        self.m_image_out_port.add_history('FlatCalibrationModule', history)
        self.m_image_out_port.copy_attributes(self.m_image_in_port)
        self.m_image_out_port.close_port()
